#if 0 //To compile, type: nmake poly2vox.c
poly2vox.exe: poly2vox.obj kplib.obj poly2vox.c;
			link poly2vox.obj kplib.obj /opt:nowin98
	del poly2vox.obj
poly2vox.obj: poly2vox.c; cl /c /J /TP poly2vox.c /Ox /Ob2 /G6Fy /MD
kplib.obj:    kplib.c;    cl /c /J /TP kplib.c    /Ox /Ob2 /G6Fy /MD
!if 0
#endif

#if 0

POLY2VOX.C by Ken Silverman (http://advsys.net/ken)

License for this code:
	* No commercial exploitation please
	* Do not remove my name or credit
	* You may distribute modified code/executables but please make it clear that it is modified

#endif

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <string.h>
#include <math.h>
#define MAX_PATH 260

#define PRINTALLOC 0
#define PRINTTIME 1

#if !defined(max)
#define max(a,b) (((a) > (b)) ? (a) : (b))
#endif
#if !defined(min)
#define min(a,b) (((a) < (b)) ? (a) : (b))
#endif

	//High-level (easy) picture loading function:
extern void kpzload (const char *, long *, long *, long *, long *);
	//ZIP functions:
extern long kzaddstack (const char *);
extern void kzuninit ();
extern long kzopen (const char *);
extern long kzread (void *, long);
extern long kzfilelength ();
extern long kzseek (long, long);
extern long kztell ();
extern long kzgetc ();
extern long kzeof ();
extern void kzclose ();
extern void kzfindfilestart (const char *);
extern long kzfindfile (char *);

#define VSID 1024   //hard-coded VXL x&y dimensions
#define MAXZDIM 256 //hard-coded VXL z dimension
#define LMAXDIM 10
#define LMAYDIM 10
#define LMAZDIM 9
#define MAXDIM (1<<LMAXDIM)
#define MAYDIM (1<<LMAYDIM)
#define MAZDIM (1<<LMAZDIM)

typedef struct { long x, y; } lpoint2d;
typedef struct { long x, y, z; } lpoint3d;
typedef struct { long x, y, z, i; } trityp;
typedef struct { long f, p, x, y; long namoff; } pictyp;
typedef struct { union { struct { float x, y, z; }; float a[3]; }; } point3d;
typedef struct { union { struct { float x, y, z; }; float a[3]; }; float u, v; } vertyp;

	//Starting maximums: bytes per unit:
long maxobjects =  128; // 8 (    1KB)
long maxnamsiz =  4096; // 1 (    4KB)
long maxmat =      512; // 8 (    4KB)
long maxpics =     256; //20 (    5KB)
long maxtris =   32768; //16 (  512KB)
long maxverts =  32768; //20 (  640KB)
long maxhash = 1048576; //12 (12288KB)

	//Dynamic arrays:
long *trioff = 0, *vertoff = 0;
vertyp *vert = 0;
trityp *tri = 0;
pictyp *pic = 0;
char *nambuf = 0; //NOTE:used for both mat&pic names
long *matstr = 0;
long *colrgb = 0;
lpoint3d *hashdat = 0; //x=indexback, y=voxelindex, z=triindex

long numobjects, numverts, numtris, numpics = 0, namplc = 0, matnum;

	//>poly2vox 3ds\\canyon.3ds canyon.vxl
	//LHASHEAD: load&rend:    all:    hashead:
	//   15:   102.620e9cc 233.270e9cc 128KB
	//   16:    47.191e9cc             256KB
	//   17:    22.054e9cc             512KB
	//   18:    13.630e9cc  42.543e9cc   1MB
	//   19:    11.138e9cc               2MB
	//   20:     9.386e9cc  24.268e9cc   4MB
	//   21:     9.508e9cc               8MB
	//   22:     8.633e9cc              16MB
	//   23:     8.658e9cc              32MB
#define LHASHEAD 20
long hashead[1<<LHASHEAD], hashcnt = 0;

#define MAXCSIZ 1028
char tempbuf[max(4096,MAXCSIZ)];

long xsiz, ysiz, zsiz, rendermode = 0, savetype;
float lastx[MAYDIM], xpiv, ypiv, zpiv;

char palfilnam[MAX_PATH] = "";

long min0[MAYDIM], max0[MAYDIM]; //MAXY
long min1[MAXDIM], max1[MAXDIM]; //MAXX
long min2[MAYDIM], max2[MAYDIM]; //MAXY

long xstart[MAXDIM+1];
unsigned short xyoffs[MAXDIM][MAYDIM+1]; //2MB

long vbit[(MAXDIM*MAYDIM*MAZDIM)>>5]; //32MB

void quitout (char *str)
{
	if (str) puts(str);

	if (hashdat) free(hashdat);
	if (colrgb) free(colrgb);
	if (matstr) free(matstr);
	if (nambuf) free(nambuf);
	if (pic) free(pic);
	if (tri) free(tri);
	if (vert) free(vert);
	if (vertoff) free(vertoff);
	if (trioff) free(trioff);
	exit(0);
}

static _inline void ftol (float f, long *a)
{
	_asm
	{
		mov eax, a
		fld f
		fistp dword ptr [eax]
	}
}

static _inline void fcossin (float a, float *c, float *s)
{
	_asm
	{
		fld a
		fsincos
		mov eax, c
		fstp dword ptr [eax]
		mov eax, s
		fstp dword ptr [eax]
	}
}

void inithash () { memset(hashead,-1,sizeof(hashead)); hashcnt = 0; }

long getval (long i)
{
#if (LHASHEAD == 20)
	long j = ((i>>8) & ((1<<LHASHEAD)-1));
#else
	long j = (((i>>LHASHEAD)-i) & ((1<<LHASHEAD)-1));
#endif
	for(j=hashead[j];j>=0;j=hashdat[j].x)
		if (hashdat[j].y == i) return(hashdat[j].z);
	return(-1);
}

void setval (long i, long v)
{
	long k;
#if (LHASHEAD == 20)
	long j = ((i>>8) & ((1<<LHASHEAD)-1));
#else
	long j = (((i>>LHASHEAD)-i) & ((1<<LHASHEAD)-1));
#endif
	for(k=hashead[j];k>=0;k=hashdat[k].x)
		if (hashdat[k].y == i) { hashdat[k].z = v; return; }
	if (hashcnt >= maxhash)
	{
		maxhash += (maxhash>>3);
#if (PRINTALLOC == 1)
		printf("maxhash=%d (%dMB)\n",maxhash,(maxhash*12)>>20);
#endif
		if (!(hashdat = (lpoint3d *)realloc(hashdat,sizeof(lpoint3d)*maxhash))) quitout("realloc failed: hashdat");
	}
	hashdat[hashcnt].x = hashead[j]; hashead[j] = hashcnt;
	hashdat[hashcnt].y = i;
	hashdat[hashcnt].z = v;
	hashcnt++;
}

static long picsloaded = 0;
static long palfreq[64][64][64];
static char rpal[256], gpal[256], bpal[256];
static long rdist[129], gdist[129], bdist[129];
static char closestcol[64][64][64];
unsigned char getclosestcol (long c32, long searchnum)
{
	long i, r, g, b, dist, mindist, *rlookup, *glookup, *blookup;
	unsigned char col, bestcol;

	if (!picsloaded) return(c32);

	r = ((c32>>16)&255); g = ((c32>>8)&255); b = (c32&255);
	col = closestcol[r>>2][g>>2][b>>2]; if (col != 255) return(col);

	rlookup = (long *)&rdist[64-(r>>2)];
	glookup = (long *)&gdist[64-(g>>2)];
	blookup = (long *)&bdist[64-(b>>2)];

	mindist = 0x7fffffff;
	for(i=0;i<searchnum;i++)
	{
		dist  = glookup[gpal[i]>>2]; if (dist >= mindist) continue;
		dist += rlookup[rpal[i]>>2]; if (dist >= mindist) continue;
		dist += blookup[bpal[i]>>2]; if (dist >= mindist) continue;
		mindist = dist; bestcol = (unsigned char)i;
	}
	closestcol[r>>2][g>>2][b>>2] = bestcol;
	return(bestcol);
}

static _inline long mulshr9 (long a, long d)
{
	_asm
	{
		mov eax, a
		mul d
		shrd eax, edx, 9
	}
}

void genpal ()
{
	unsigned short *s;
	long i, j, c, r, g, b, p, v, x, y, z, xx, yy, zz, xxx, yyy, zzz, *dc;

	j = 0;
	for(i=64;i>=0;i--)
	{
		//j = (i-64)*(i-64);
		rdist[i] = rdist[128-i] = j*30;
		gdist[i] = gdist[128-i] = j*59;
		bdist[i] = bdist[128-i] = j*11;
		j += 129-(i<<1);
	}

	memset(closestcol,-1,sizeof(closestcol));
	c = (rand()<<17);
	for(i=0;i<256;i++)
	{
		if (!i)
			{ rpal[i] = gpal[i] = bpal[i] = 0; }
		else
		{
			v = 0;
			for(j=1024;j>0;j--)
			{
				c = (c*27584621)+1;
				dc = &palfreq[0][0][((unsigned)c)>>14];
				p = dc[0];
				r = ((c>>26)&63); if (r !=  0) p += dc[-64*64];
										if (r != 63) p += dc[+64*64];
				g = ((c>>20)&63); if (g !=  0) p += dc[-   64];
										if (g != 63) p += dc[+   64];
				b = ((c>>14)&63); if (b !=  0) p += dc[-    1];
										if (b != 63) p += dc[+    1];
				if (p > v) { v = p; rpal[i] = (r<<2); gpal[i] = (g<<2); bpal[i] = (b<<2); }
			}
		}

		xx = max((rpal[i]>>2)-2,0); xxx = min((rpal[i]>>2)+2,63);
		yy = max((gpal[i]>>2)-2,0); yyy = min((gpal[i]>>2)+2,63);
		zz = max((bpal[i]>>2)-2,0); zzz = min((bpal[i]>>2)+2,63);
		for(g=yy;g<=yyy;g++)
		{
			y = gdist[(gpal[i]>>2)-g+64]; if (y >= 512) continue;
			for(r=xx;r<=xxx;r++)
			{
				x = rdist[(rpal[i]>>2)-r+64]+y; if (x >= 512) continue;
				for(b=zz;b<=zzz;b++)
				{
					z = bdist[(bpal[i]>>2)-b+64]+x; if (z >= 512) continue;
					palfreq[r][g][b] = mulshr9(palfreq[r][g][b],z);
				}
			}
		}
	}
}

	//Fancy interpolation filter for 4x4 matrix (256 samples, 4 coefficients)
	//NOTE: this is higher quality than bilinear filter (which uses 2x2 matrix)
static short lut4x4[256*4];
static long lut4x4inited = 0;
static float calclut4x4 (float x)
{
	x = fabs(x); //4x4 filter (spline16)
	if (x < 1) return(((x-1.8)*x - 0.2)*x + 1.0);
	x -= 1.0; return(((x*(-1.0/3.0) + 0.8)*x - (7.0/15.0))*x);
}
static void initlut4x4 ()
{
	long i, k;
	float f;

	lut4x4inited = 1;
	for(i=0;i<256;i++)
	{
		f = ((float)i) / 256.0;
		ftol(calclut4x4(f-2.0)*-32768.0,&k); lut4x4[i*4+3] = k;
		ftol(calclut4x4(f-1.0)*-32768.0,&k); lut4x4[i*4+2] = k;
		ftol(calclut4x4(f+0.0)*-32768.0,&k); lut4x4[i*4+1] = k;
		ftol(calclut4x4(f+1.0)*-32768.0,&k); lut4x4[i*4+0] = k;
	}
}

long getpix32 (long x, long y, long z)
{
	float fx, fy, fz, fx1, fy1, fz1, fx2, fy2, fz2, fu, fv;
	float f, nx, ny, nz, ux, uy, uz, vx, vy, vz;
	long i, j, k, u, v, tu, tv, uu, vv, wu, wv, r, g, b, rr, gg, bb;

	i = getval((x*MAYDIM+y)*MAZDIM+z); if (i < 0) return(0x808080);
	u = tri[i].i; j = ~colrgb[u];
	if ((colrgb[u] >= 0) || (!pic[j].f))
	{
		if ((!picsloaded) && (savetype < 2)) return(u);
		return(colrgb[u]); //solid color
	}

	if ((pic[j].x <= 0) || (pic[j].y <= 0)) return(0);
	fx = vert[tri[i].x].x; fx1 = vert[tri[i].y].x-fx; fx2 = vert[tri[i].z].x-fx;
	fy = vert[tri[i].x].y; fy1 = vert[tri[i].y].y-fy; fy2 = vert[tri[i].z].y-fy;
	fz = vert[tri[i].x].z; fz1 = vert[tri[i].y].z-fz; fz2 = vert[tri[i].z].z-fz;
	nx = fy1*fz2 - fz1*fy2;
	ny = fz1*fx2 - fx1*fz2;
	nz = fx1*fy2 - fy1*fx2;

		//Cramer's rule derivation:
		//u(&vert[tri[i].x]) = 0.0, v(&vert[tri[i].x]) = 0.0
		//u(&vert[tri[i].y]) = 1.0, v(&vert[tri[i].y]) = 0.0
		//u(&vert[tri[i].z]) = 0.0, v(&vert[tri[i].z]) = 1.0
		//fx1*ux + fy1*uy + fz1*uz = 1    fx1*vx + fy1*vy + fz1*vz = 0
		//fx2*ux + fy2*uy + fz2*uz = 0    fx2*vx + fy2*vy + fz2*vz = 1
		// nx*ux +  ny*uy +  nz*uz = 0     nx*vx +  ny*vy +  nz*vz = 0
	f = fx1*(fy2*nz-fz2*ny) + fy1*(fz2*nx-fx2*nz) + fz1*(fx2*ny-fy2*nx);
	if (f != 0)
	{
		f = 1.0 / f;
		ux = (fy2*nz-fz2*ny)*f; uy = (fz2*nx-fx2*nz)*f; uz = (fx2*ny-fy2*nx)*f;
		vx = (fz1*ny-fy1*nz)*f; vy = (fx1*nz-fz1*nx)*f; vz = (fy1*nx-fx1*ny)*f;
	} else { ux = uy = uz = vx = vy = vz = 0; }
	fu = ((float)x-fx)*ux + ((float)y-fy)*uy + ((float)z-fz)*uz;
	fv = ((float)x-fx)*vx + ((float)y-fy)*vy + ((float)z-fz)*vz;
	fx = (vert[tri[i].y].u-vert[tri[i].x].u)*fu +
		  (vert[tri[i].z].u-vert[tri[i].x].u)*fv + vert[tri[i].x].u;
	fy = (vert[tri[i].y].v-vert[tri[i].x].v)*fu +
		  (vert[tri[i].z].v-vert[tri[i].x].v)*fv + vert[tri[i].x].v;
#if 0
		//Nearest texture filter
	ftol(((float)pic[j].x)*fx,&u);
	ftol(((float)pic[j].y)*fy,&v);
	if (u < 0) u += ((pic[j].x-1-u)/pic[j].x)*pic[j].x; else u %= pic[j].x;
	if (v < 0) v += ((pic[j].y-1-v)/pic[j].y)*pic[j].y; else v %= pic[j].y;
	j = *(long *)((pic[j].y-1-v)*pic[j].p + (u<<2) + pic[j].f);
	return(j);
#else
		//Fancy 4x4 texture filter
	ftol(((float)(pic[j].x<<8))*fx,&u); u -= 384;
	ftol(((float)(pic[j].y<<8))*fy,&v); v += 384;
	if (u < 0) u += (((pic[j].x<<8)-1-u)/(pic[j].x<<8))*(pic[j].x<<8); else u %= (pic[j].x<<8);
	if (v < 0) v += (((pic[j].y<<8)-1-v)/(pic[j].y<<8))*(pic[j].y<<8); else v %= (pic[j].y<<8);
	v = (pic[j].y<<8)-1-v;

	if (!lut4x4inited) initlut4x4();
	r = g = b = 0;
	for(vv=0,tv=(v>>8);vv<4;vv++)
	{
		rr = gg = bb = 0;
		for(uu=0,tu=(u>>8);uu<4;uu++)
		{
			k = (*(long *)(tv*pic[j].p + (tu<<2) + pic[j].f));
			wu = lut4x4[(u&255)*4+uu];
			rr += ((((k>>16)&255)*wu)>>8);
			gg += ((((k>> 8)&255)*wu)>>8);
			bb += ((( k     &255)*wu)>>8);
			tu++; if (tu >= pic[j].x) tu = 0;
		}
		wv = lut4x4[(v&255)*4+vv];
		r += ((rr*wv)>>8);
		g += ((gg*wv)>>8);
		b += ((bb*wv)>>8);
		tv++; if (tv >= pic[j].y) tv = 0;
	}
	r = min(max(r>>14,0),255);
	g = min(max(g>>14,0),255);
	b = min(max(b>>14,0),255);
	return((r<<16)+(g<<8)+b);
#endif
}

//EQUIVEC code begins -----------------------------------------------------

#define PI 3.141592653589793
#define GOLDRAT 0.3819660112501052 //Golden Ratio: 1 - 1/((sqrt(5)+1)/2)
typedef struct
{
	float fibx[45], fiby[45];
	float azval[20], zmulk, zaddk;
	long fib[47], aztop, npoints;
	point3d *p;  //For memory version :/
	long pcur;
} equivectyp;
static equivectyp equivec;

static _inline long dmulshr0 (long a, long d, long s, long t)
{
	_asm
	{
		mov eax, a
		imul d
		mov ecx, eax
		mov eax, s
		imul t
		add eax, ecx
	}
}

void equiind2vec (long i, float *x, float *y, float *z)
{
	float r;
	(*z) = (float)i*equivec.zmulk + equivec.zaddk; r = sqrt(1.f - (*z)*(*z));
	fcossin((float)i*(GOLDRAT*PI*2),x,y); (*x) *= r; (*y) *= r;
}

	//(Pass n=0 to free buffer)
long equimemset (long n)
{
	long z;

	if (equivec.pcur == n) return(1); //Don't recalculate if same #
	if (equivec.p) { free(equivec.p); equivec.p = 0; }
	if (!n) return(1);

		//Init vector positions (equivec.p) for memory version
	if (!(equivec.p = (point3d *)malloc(((n+1)&~1)*sizeof(point3d))))
		return(0);

	equivec.pcur = n;
	equivec.zmulk = 2 / (float)n; equivec.zaddk = equivec.zmulk*.5 - 1.0;
	for(z=n-1;z>=0;z--)
		equiind2vec(z,&equivec.p[z].x,&equivec.p[z].y,&equivec.p[z].z);
	if (n&1) //Hack for when n=255 and want a <0,0,0> vector
		{ equivec.p[n].x = equivec.p[n].y = equivec.p[n].z = 0; }
	return(1);
}

	//Very fast; good quality, requires equivec.p[] :/
long equivec2indmem (float x, float y, float z)
{
	long b, i, j, k, bestc;
	float xy, zz, md, d;

	xy = atan2(y,x); //atan2 is 150 clock cycles!
	j = ((*(long *)&z)&0x7fffffff);
	bestc = equivec.aztop;
	do
	{
		if (j < *(long *)&equivec.azval[bestc]) break;
		bestc--;
	} while (bestc);

	zz = z + 1.f;
	ftol(equivec.fibx[bestc]*xy + equivec.fiby[bestc]*zz - .5,&i);
	bestc++;
	ftol(equivec.fibx[bestc]*xy + equivec.fiby[bestc]*zz - .5,&j);

	k = dmulshr0(equivec.fib[bestc+2],i,equivec.fib[bestc+1],j);
	if ((unsigned long)k < equivec.npoints)
	{
		md = equivec.p[k].x*x + equivec.p[k].y*y + equivec.p[k].z*z;
		j = k;
	} else md = -2.f;
	b = bestc+3;
	do
	{
		i = equivec.fib[b] + k;
		if ((unsigned long)i < equivec.npoints)
		{
			d = equivec.p[i].x*x + equivec.p[i].y*y + equivec.p[i].z*z;
			if (*(long *)&d > *(long *)&md) { md = d; j = i; }
		}
		b--;
	} while (b != bestc);
	return(j);
}

void equivecinit (long n)
{
	float t0, t1;
	long z;

		//Init constants for ind2vec
	equivec.npoints = n;
	equivec.zmulk = 2 / (float)n; equivec.zaddk = equivec.zmulk*.5 - 1.0;

	equimemset(n);

		//Init Fibonacci table
	equivec.fib[0] = 0; equivec.fib[1] = 1;
	for(z=2;z<47;z++) equivec.fib[z] = equivec.fib[z-2]+equivec.fib[z-1];

		//Init fibx/y LUT
	t0 = .5 / PI; t1 = (float)n * -.5;
	for(z=0;z<45;z++)
	{
		t0 = -t0; equivec.fibx[z] = (float)equivec.fib[z+2]*t0;
		t1 = -t1; equivec.fiby[z] = ((float)equivec.fib[z+2]*GOLDRAT - (float)equivec.fib[z])*t1;
	}

	t0 = 1 / ((float)n * PI);
	for(equivec.aztop=0;equivec.aztop<20;equivec.aztop++)
	{
		t1 = 1 - (float)equivec.fib[(equivec.aztop<<1)+6]*t0; if (t1 < 0) break;
		equivec.azval[equivec.aztop+1] = sqrt(t1);
	}
}

void equivecuninit () { equimemset(0); }

//EQUIVEC code ends -------------------------------------------------------

long getnorm8 (long x, long y, long z)
{
	float fx, fy, fz, fx1, fy1, fz1, fx2, fy2, fz2, f;
	long i;

	i = getval((x*MAYDIM+y)*MAZDIM+z); if (i < 0) return(0);

	fx = vert[tri[i].x].x; fx1 = vert[tri[i].y].x-fx; fx2 = vert[tri[i].z].x-fx;
	fy = vert[tri[i].x].y; fy1 = vert[tri[i].y].y-fy; fy2 = vert[tri[i].z].y-fy;
	fz = vert[tri[i].x].z; fz1 = vert[tri[i].y].z-fz; fz2 = vert[tri[i].z].z-fz;
	fx = fz1*fy2 - fy1*fz2; //Note: inverted from normal!
	fy = fx1*fz2 - fz1*fx2;
	fz = fy1*fx2 - fx1*fy2;
	f = sqrt(fx*fx + fy*fy + fz*fz); if (f > 0) f = 1.0 / f;
		//NOTE: don't use rsqrtss: sqrt must be accurate here
	return(equivec2indmem(fx*f,fy*f,fz*f));
}

//FLOODFILL2D/3D begins -----------------------------------------------------

#define FILLBUFSIZ 16384 //Use realloc instead!
typedef struct { unsigned short x, y, z0, z1; } cpoint4d; //128K
static cpoint4d fbuf[FILLBUFSIZ];

static __forceinline long bsf (long a) { _asm bsf eax, a }
static __forceinline long bsr (long a) { _asm bsr eax, a }

static __forceinline long dntil0 (long *lptr, long z)
{
	//   //This line does the same thing (but slow & brute force)
	//while ((z < zsiz) && (lptr[z>>5]&(1<<z))) z++; return(z);
	long i;
		//WARNING: zsiz must be multiple of 32!
	i = (lptr[z>>5]|((1<<z)-1)); z &= ~31;
	while (i == -1)
	{
		z += 32; if (z >= zsiz) return(zsiz);
		i = lptr[z>>5];
	}
	return(bsf(~i)+z);

	//                 z=11
	//                    |
	//11110101110000100111111110011111 lptr[z>>5]
	//00000000000000000000011111111111 (1<<z)-1
	//11110101110000100111111111111111
}

static __forceinline long dntil1 (long *lptr, long z)
{
	//   //This line does the same thing (but slow & brute force)
	//while ((z < zsiz) && (!(lptr[z>>5]&(1<<z)))) z++; return(z);
	long i;
		//WARNING: zsiz must be multiple of 32!
	i = (lptr[z>>5]&(-1<<z)); z &= ~31;
	while (!i)
	{
		z += 32; if (z >= zsiz) return(zsiz);
		i = lptr[z>>5];
	}
	return(bsf(i)+z);
}

static __forceinline long uptil1 (long *lptr, long z)
{
	//   //This line does the same thing (but slow & brute force)
	//while ((z > 0) && (!(lptr[(z-1)>>5]&(1<<(z-1))))) z--; return(z);
	long i;
		//WARNING: zsiz must be multiple of 32!
	if (!z) return(0); //Prevent possible crash
	i = (lptr[(z-1)>>5]&((1<<z)-1)); z &= ~31;
	while (!i)
	{
		z -= 32; if (z < 0) return(0);
		i = lptr[z>>5];
	}
	return(bsr(i)+z+1);
}

	//Set all bits in vbit from (x,y,z0) to (x,y,z1-1) to 0's
static __forceinline void setzrange0 (long *lptr, long z0, long z1)
{
	long z, ze;
	if (!((z0^z1)&~31)) { lptr[z0>>5] &= ((~(-1<<z0))|(-1<<z1)); return; }
	z = (z0>>5); ze = (z1>>5);
	lptr[z] &=~(-1<<z0); for(z++;z<ze;z++) lptr[z] = 0;
	lptr[z] &= (-1<<z1);
}

	//Set all bits in vbit from (x,y,z0) to (x,y,z1-1) to 1's
static __forceinline void setzrange1 (long *lptr, long z0, long z1)
{
	long z, ze;
	if (!((z0^z1)&~31)) { lptr[z0>>5] |= ((~(-1<<z1))&(-1<<z0)); return; }
	z = (z0>>5); ze = (z1>>5);
	lptr[z] |= (-1<<z0); for(z++;z<ze;z++) lptr[z] = -1;
	lptr[z] |=~(-1<<z1);
}

void floodfill3dbits (long x, long y, long z) //Conducts on 0's and sets to 1's
{
	long i, j, xy, z0, z1, i0, i1;
	cpoint4d a;

	xy = (x*MAYDIM+y)*MAZDIM;
	if (vbit[(xy+z)>>5]&(1<<z)) return;
	a.x = x; a.z0 = uptil1(&vbit[xy>>5],z);
	a.y = y; a.z1 = dntil1(&vbit[xy>>5],z+1);
	setzrange1(&vbit[xy>>5],a.z0,a.z1);
	i0 = i1 = 0; goto floodfill3dskip;
	do
	{
		a = fbuf[i0]; i0 = ((i0+1)&(FILLBUFSIZ-1));
floodfill3dskip:;
		for(j=3;j>=0;j--)
		{
			if (j&1) { x = a.x+(j&2)-1; if ((unsigned long)x >= xsiz) continue; y = a.y; }
				 else { y = a.y+(j&2)-1; if ((unsigned long)y >= ysiz) continue; x = a.x; }
			xy = (x*MAYDIM+y)*MAZDIM;

			if (vbit[(xy+a.z0)>>5]&(1<<a.z0)) { z0 = dntil0(&vbit[xy>>5],a.z0); z1 = z0; }
												  else { z0 = uptil1(&vbit[xy>>5],a.z0); z1 = a.z0; }
			while (z1 < a.z1)
			{
				z1 = dntil1(&vbit[xy>>5],z1);
				fbuf[i1].x = x; fbuf[i1].y = y;
				fbuf[i1].z0 = z0; fbuf[i1].z1 = z1;
				i1 = ((i1+1)&(FILLBUFSIZ-1));
				if (i0 == i1) puts("WARNING: floodfill3d stack overflow!");
				setzrange1(&vbit[xy>>5],z0,z1);
				z0 = dntil0(&vbit[xy>>5],z1); z1 = z0;
			}
		}
	} while (i0 != i1);
}

//FLOODFILL2D/3D ends -------------------------------------------------------

void renderfilltri (long faceind)
{
	long i, j, ind[3], sy[3], sx0, sx1, x, y, z;
	float f, u, fx, fy, fz, fx1, fy1, fz1, fx2, fy2, fz2;

	ind[0] = tri[faceind].x; ind[1] = tri[faceind].y; ind[2] = tri[faceind].z;

	fx = vert[ind[0]].x; fx1 = vert[ind[1]].x-fx; fx2 = vert[ind[2]].x-fx;
	fy = vert[ind[0]].y; fy1 = vert[ind[1]].y-fy; fy2 = vert[ind[2]].y-fy;
	fz = vert[ind[0]].z; fz1 = vert[ind[1]].z-fz; fz2 = vert[ind[2]].z-fz;
	f = 1.0f/(fx1*fy2 - fy1*fx2);
	fx1 = (fx1*fz2 - fz1*fx2)*f;
	fy1 = (fz1*fy2 - fy1*fz2)*f;
	fz -= (fy1*fx + fx1*fy);

	for(i=2;i>=0;i--) sy[i] = (long)vert[ind[i]].y;
	for(i=2,j=0;i>=0;j=i--)
	{
		if (sy[j] >= sy[i]) continue;
		f = (vert[ind[i]].x-vert[ind[j]].x)/(vert[ind[i]].y-vert[ind[j]].y);
		for(y=sy[j]+1;y<=sy[i];y++)
			lastx[y] = ((float)y-vert[ind[j]].y)*f + vert[ind[j]].x;
	}
	for(i=2,j=0;i>=0;j=i--)
	{
		if (sy[j] <= sy[i]) continue;
		f = (vert[ind[j]].x-vert[ind[i]].x)/(vert[ind[j]].y-vert[ind[i]].y);
		for(y=sy[i]+1;y<=sy[j];y++)
		{
			u = ((float)y-vert[ind[i]].y)*f + vert[ind[i]].x;
			sx0 = (long)min(lastx[y],u);
			sx1 = (long)max(lastx[y],u);
			for(x=sx0+1;x<=sx1;x++)
			{
				z = (x*MAYDIM + y)*MAZDIM + ((long)(x*fy1 + y*fx1 + fz));
				vbit[z>>5] ^= (1<<z);
				setval(z,faceind);
			}
		}
	}
}

void renderfillfinish ()
{
	long i, j, x, y, z;

		//convert xor-style geometry to solid geometry
	for(x=0;x<xsiz;x++)
		for(y=0;y<ysiz;y++)
		{
			j = (x*MAYDIM + y)*MAZDIM; z = -1;
			do
			{
				i = dntil1(&vbit[j>>5],z+1); if (i >= zsiz) break;
				z = dntil1(&vbit[j>>5],i+1);
				setzrange1(&vbit[j>>5],i+1,z); if (z >= zsiz) break;
				vbit[(j+z)>>5] &= ~(1<<z);
			} while (z < zsiz);
		}
}

void canseerange (point3d *p0, point3d *p1)
{
	lpoint3d a, c, d, p, i;
	point3d f, g;
	long cnt;

	ftol(p0->x-.5f,&a.x); ftol(p0->y-.5f,&a.y); ftol(p0->z-.5f,&a.z);
	ftol(p1->x-.5f,&c.x); ftol(p1->y-.5f,&c.y); ftol(p1->z-.5f,&c.z);
	cnt = 0;

		  if (c.x <  a.x) { d.x = -1; f.x = p0->x-a.x;   g.x = (p0->x-p1->x)*1024; cnt += a.x-c.x; }
	else if (c.x != a.x) { d.x =  1; f.x = a.x+1-p0->x; g.x = (p1->x-p0->x)*1024; cnt += c.x-a.x; }
	else f.x = g.x = 0;
		  if (c.y <  a.y) { d.y = -1; f.y = p0->y-a.y;   g.y = (p0->y-p1->y)*1024; cnt += a.y-c.y; }
	else if (c.y != a.y) { d.y =  1; f.y = a.y+1-p0->y; g.y = (p1->y-p0->y)*1024; cnt += c.y-a.y; }
	else f.y = g.y = 0;
		  if (c.z <  a.z) { d.z = -1; f.z = p0->z-a.z;   g.z = (p0->z-p1->z)*1024; cnt += a.z-c.z; }
	else if (c.z != a.z) { d.z =  1; f.z = a.z+1-p0->z; g.z = (p1->z-p0->z)*1024; cnt += c.z-a.z; }
	else f.z = g.z = 0;

	ftol(f.x*g.z - f.z*g.x,&p.x); ftol(g.x,&i.x);
	ftol(f.y*g.z - f.z*g.y,&p.y); ftol(g.y,&i.y);
	ftol(f.y*g.x - f.x*g.y,&p.z); ftol(g.z,&i.z);
	for(;cnt;cnt--)
	{
			//use a.x, a.y, a.z
		if (a.x < min0[a.y]) min0[a.y] = a.x;
		if (a.x > max0[a.y]) max0[a.y] = a.x;
		if (a.z < min1[a.x]) min1[a.x] = a.z;
		if (a.z > max1[a.x]) max1[a.x] = a.z;
		if (a.z < min2[a.y]) min2[a.y] = a.z;
		if (a.z > max2[a.y]) max2[a.y] = a.z;

		if (((p.x|p.y) >= 0) && (a.z != c.z)) { a.z += d.z; p.x -= i.x; p.y -= i.y; }
		else if ((p.z >= 0) && (a.x != c.x))  { a.x += d.x; p.x += i.z; p.z -= i.y; }
		else                                  { a.y += d.y; p.y += i.z; p.z += i.x; }
	}
}

void renderface1tri (long faceind)
{
	point3d n, *p0, *p1, *p2;
	float f, x0, x1, y0, y1, rx, ry, k0, k1;
	long i, x, y, z, z0, z1, minx, maxx, miny, maxy;

	p0 = (point3d *)&vert[tri[faceind].x];
	p1 = (point3d *)&vert[tri[faceind].y];
	p2 = (point3d *)&vert[tri[faceind].z];

	x0 = p0->x; x1 = p1->x; if (x0 > x1) { f = x0; x0 = x1; x1 = f; }
	if (p2->x < x0) x0 = p2->x;
	if (p2->x > x1) x1 = p2->x;
	y0 = p0->y; y1 = p1->y; if (y0 > y1) { f = y0; y0 = y1; y1 = f; }
	if (p2->y < y0) y0 = p2->y;
	if (p2->y > y1) y1 = p2->y;
	ftol(x0-.5f,&minx); ftol(y0-.5f,&miny);
	ftol(x1-.5f,&maxx); ftol(y1-.5f,&maxy);
	for(i=miny;i<=maxy;i++) { min0[i] = 0x7fffffff; max0[i] = 0x80000000; }
	for(i=minx;i<=maxx;i++) { min1[i] = 0x7fffffff; max1[i] = 0x80000000; }
	for(i=miny;i<=maxy;i++) { min2[i] = 0x7fffffff; max2[i] = 0x80000000; }

	canseerange(p0,p1);
	canseerange(p1,p2);
	canseerange(p2,p0);

	n.x = (p1->z-p0->z)*(p2->y-p1->y) - (p1->y-p0->y) * (p2->z-p1->z);
	n.y = (p1->x-p0->x)*(p2->z-p1->z) - (p1->z-p0->z) * (p2->x-p1->x);
	n.z = (p1->y-p0->y)*(p2->x-p1->x) - (p1->x-p0->x) * (p2->y-p1->y);
	f = 1.0f / (float)sqrt(n.x*n.x + n.y*n.y + n.z*n.z); if (n.z < 0.f) f = -f;
	n.x *= f; n.y *= f; n.z *= f;

		//Change this to "if (0)" for cool bump-mapping effects on kerolamp&skull
	if (n.z > .01f)
	{
		f = -1.0f / n.z; rx = n.x*f; ry = n.y*f;
		k0 = ((n.x>=0)-p0->x)*rx + ((n.y>=0)-p0->y)*ry - ((n.z>=0)-p0->z) + .5f;
		k1 = ((n.x< 0)-p0->x)*rx + ((n.y< 0)-p0->y)*ry - ((n.z< 0)-p0->z) - .5f;
	}
	else { rx = 0.f; ry = 0.f; k0 = -2147000000.0f; k1 = 2147000000.0f; }

	for(y=miny;y<=maxy;y++)
		for(x=min0[y];x<=max0[y];x++)
		{
			f = (float)x*rx + (float)y*ry; ftol(f+k0,&z0); ftol(f+k1,&z1);
			if (z0 < min1[x]) z0 = min1[x];
			if (z1 > max1[x]) z1 = max1[x];
			if (z0 < min2[y]) z0 = min2[y];
			if (z1 > max2[y]) z1 = max2[y];

			i = (x*MAYDIM+y)*MAZDIM;
			for(z=z0;z<=z1;z++) //setcube(x,y,z);
			{
				vbit[(i+z)>>5] |= (1<<(i+z));
				setval(i+z,faceind);
			}
		}
}

	//Minimal voxel implementation
void renderline0 (point3d *bp0, point3d *bp1, long col)
{
	float dx, dy, dz;
	point3d p0, p1;
	lpoint3d l0, l1, l2;
	long j;

	p0 = *bp0; p1 = *bp1;

	ftol(p0.x+.5f,&l0.x); ftol(p0.y+.5f,&l0.y); ftol(p0.z+.5f,&l0.z);
	ftol(p1.x+.5f,&l1.x); ftol(p1.y+.5f,&l1.y); ftol(p1.z+.5f,&l1.z);

	dx = (p1.x)-(p0.x); dy = (p1.y)-(p0.y); dz = (p1.z)-(p0.z);
	if ((fabs(dz) >= fabs(dx)) && (fabs(dz) >= fabs(dy)))
	{
		if (l0.z > l1.z) { l2 = l0; l0 = l1; l1 = l2; p0 = p1; }
		dz = 1.0f / dz; dx *= dz; dy *= dz; p0.x -= p0.z*dx; p0.y -= p0.z*dy;
		for(l2.z=l0.z;l2.z<l1.z;l2.z++)
		{
			ftol(((float)l2.z)*dx + p0.x,&l2.x);
			ftol(((float)l2.z)*dy + p0.y,&l2.y);
			j = (l2.x*MAYDIM + l2.y)*MAZDIM + l2.z; vbit[j>>5] |= (1<<j);
			setval(j,col);
		}
	}
	else if (fabs(dx) >= fabs(dy))
	{
		if (l0.x > l1.x) { l2 = l0; l0 = l1; l1 = l2; p0 = p1; }
		dx = 1.0f / dx; dy *= dx; dz *= dx; p0.y -= p0.x*dy; p0.z -= p0.x*dz;
		for(l2.x=l0.x;l2.x<l1.x;l2.x++)
		{
			ftol(((float)l2.x)*dy + p0.y,&l2.y);
			ftol(((float)l2.x)*dz + p0.z,&l2.z);
			j = (l2.x*MAYDIM + l2.y)*MAZDIM + l2.z; vbit[j>>5] |= (1<<j);
			setval(j,col);
		}
	}
	else
	{
		if (l0.y > l1.y) { l2 = l0; l0 = l1; l1 = l2; p0 = p1; }
		dy = 1.0f / dy; dx *= dy; dz *= dy; p0.x -= p0.y*dx; p0.z -= p0.y*dz;
		for(l2.y=l0.y;l2.y<l1.y;l2.y++)
		{
			ftol(((float)l2.y)*dx + p0.x,&l2.x);
			ftol(((float)l2.y)*dz + p0.z,&l2.z);
			j = (l2.x*MAYDIM + l2.y)*MAZDIM + l2.z; vbit[j>>5] |= (1<<j);
			setval(j,col);
		}
	}
}

void renderline0tri (long i)
{
	renderline0((point3d *)&vert[tri[i].x],(point3d *)&vert[tri[i].y],i);
	renderline0((point3d *)&vert[tri[i].y],(point3d *)&vert[tri[i].z],i);
	renderline0((point3d *)&vert[tri[i].z],(point3d *)&vert[tri[i].x],i);
}

	//Based on cansee from voxlap5.c (Supercover)
void renderline1 (point3d *p0, point3d *p1, long col)
{
	lpoint3d a, c, d, p, i;
	point3d f, g;
	long cnt, j;

	ftol(p0->x-.5f,&a.x); ftol(p0->y-.5f,&a.y); ftol(p0->z-.5f,&a.z);
	j = (a.x*MAYDIM + a.y)*MAZDIM + a.z; vbit[j>>5] |= (1<<j);
	setval(j,col);
	ftol(p1->x-.5f,&c.x); ftol(p1->y-.5f,&c.y); ftol(p1->z-.5f,&c.z);
	cnt = 0;

		  if (c.x <  a.x) { d.x = -1; f.x = p0->x-a.x;   g.x = (p0->x-p1->x)*1024; cnt += a.x-c.x; }
	else if (c.x != a.x) { d.x =  1; f.x = a.x+1-p0->x; g.x = (p1->x-p0->x)*1024; cnt += c.x-a.x; }
	else f.x = g.x = 0;
		  if (c.y <  a.y) { d.y = -1; f.y = p0->y-a.y;   g.y = (p0->y-p1->y)*1024; cnt += a.y-c.y; }
	else if (c.y != a.y) { d.y =  1; f.y = a.y+1-p0->y; g.y = (p1->y-p0->y)*1024; cnt += c.y-a.y; }
	else f.y = g.y = 0;
		  if (c.z <  a.z) { d.z = -1; f.z = p0->z-a.z;   g.z = (p0->z-p1->z)*1024; cnt += a.z-c.z; }
	else if (c.z != a.z) { d.z =  1; f.z = a.z+1-p0->z; g.z = (p1->z-p0->z)*1024; cnt += c.z-a.z; }
	else f.z = g.z = 0;

	ftol(f.x*g.z - f.z*g.x,&p.x); ftol(g.x,&i.x);
	ftol(f.y*g.z - f.z*g.y,&p.y); ftol(g.y,&i.y);
	ftol(f.y*g.x - f.x*g.y,&p.z); ftol(g.z,&i.z);
	for(;cnt;cnt--)
	{
		if (((p.x|p.y) >= 0) && (a.z != c.z)) { a.z += d.z; p.x -= i.x; p.y -= i.y; }
		else if ((p.z >= 0) && (a.x != c.x))  { a.x += d.x; p.x += i.z; p.z -= i.y; }
		else                                  { a.y += d.y; p.y += i.z; p.z += i.x; }
		j = (a.x*MAYDIM + a.y)*MAZDIM + a.z; vbit[j>>5] |= (1<<j);
		setval(j,col);
	}
}

void renderline1tri (long i)
{
	renderline1((point3d *)&vert[tri[i].x],(point3d *)&vert[tri[i].y],i);
	renderline1((point3d *)&vert[tri[i].y],(point3d *)&vert[tri[i].z],i);
	renderline1((point3d *)&vert[tri[i].z],(point3d *)&vert[tri[i].x],i);
}

void renderface0tri (long faceind)
{
	float f, u, fx, fy, fz, fx1, fy1, fz1, fx2, fy2, fz2;
	long i, j, k, l, ind[3], sy[3], sx0, sx1, x, y, z, i0, i1, i2;
	char xsh, ysh, zsh;

	ind[0] = tri[faceind].x; ind[1] = tri[faceind].y; ind[2] = tri[faceind].z;

#if 0
	  //Use normal vector to calculate best axis
	fx = vert[ind[0]].x; fx1 = vert[ind[1]].x-fx; fx2 = vert[ind[2]].x-fx;
	fy = vert[ind[0]].y; fy1 = vert[ind[1]].y-fy; fy2 = vert[ind[2]].y-fy;
	fz = vert[ind[0]].z; fz1 = vert[ind[1]].z-fz; fz2 = vert[ind[2]].z-fz;
	fx = fy1*fz2 - fz1*fy2;
	fy = fz1*fx2 - fx1*fz2;
	fz = fx1*fy2 - fy1*fx2;
	if ((fabs(fz) >= fabs(fx)) && (fabs(fz) >= fabs(fy))) k = 0;
	else if (fabs(fx) >= fabs(fy)) k = 1; else k = 2;
#endif

	for(k=0;k<3;k++)
	{
		i0 = k; i2 = k-1; if (i2 < 0) i2 = 2; i1 = 3-i0-i2;
		fx = vert[ind[0]].a[i0]; fx1 = vert[ind[1]].a[i0]-fx; fx2 = vert[ind[2]].a[i0]-fx;
		fy = vert[ind[0]].a[i1]; fy1 = vert[ind[1]].a[i1]-fy; fy2 = vert[ind[2]].a[i1]-fy;
		fz = vert[ind[0]].a[i2]; fz1 = vert[ind[1]].a[i2]-fz; fz2 = vert[ind[2]].a[i2]-fz;
		f = 1.0f/(fx1*fy2 - fy1*fx2);
		fx1 = (fx1*fz2 - fz1*fx2)*f;
		fy1 = (fz1*fy2 - fy1*fz2)*f;
		fz -= (fy1*fx + fx1*fy);

		switch (k)
		{
			case 0: xsh = LMAYDIM+LMAZDIM; ysh = LMAZDIM; zsh = 0; break;
			case 1: zsh = LMAYDIM+LMAZDIM; xsh = LMAZDIM; ysh = 0; break;
			case 2: ysh = LMAYDIM+LMAZDIM; zsh = LMAZDIM; xsh = 0; break;
		}

		for(i=2;i>=0;i--) sy[i] = (long)vert[ind[i]].a[i1];
		for(i=2,j=0;i>=0;j=i--)
		{
			if (sy[j] >= sy[i]) continue;
			f = (vert[ind[i]].a[i0]-vert[ind[j]].a[i0])/(vert[ind[i]].a[i1]-vert[ind[j]].a[i1]);
			for(y=sy[j]+1;y<=sy[i];y++)
				lastx[y] = ((float)y-vert[ind[j]].a[i1])*f + vert[ind[j]].a[i0];
		}
		for(i=2,j=0;i>=0;j=i--)
		{
			if (sy[j] <= sy[i]) continue;
			f = (vert[ind[j]].a[i0]-vert[ind[i]].a[i0])/(vert[ind[j]].a[i1]-vert[ind[i]].a[i1]);
			for(y=sy[i]+1;y<=sy[j];y++)
			{
				u = ((float)y-vert[ind[i]].a[i1])*f + vert[ind[i]].a[i0];
				sx0 = (long)min(lastx[y],u);
				sx1 = (long)max(lastx[y],u);
				for(x=sx0+1;x<=sx1;x++)
				{
					z = (long)(x*fy1 + y*fx1 + fz);
					l = (x<<xsh)+(y<<ysh)+(z<<zsh); vbit[l>>5] |= (1<<l);
					setval(l,faceind);
				}
			}
		}
	}
}

//---------------------------------------------------------------------------

long findchar (char *st, char ch)
{
	long i;
	for(i=0;st[i];i++) if (st[i] == ch) return(i);
	return(i);
}

long loadasc (char *fnam)
{
	long i, j, w, x, y, z, op, p, hidden, lastface = 0;

	numobjects = numverts = numtris = numpics = 0;

		//Make material index #0 a dummy color (black)
	matstr[0] = namplc;
	strcpy(&nambuf[namplc],"r0g0b0a0"); namplc += strlen(&nambuf[namplc])+1;
	colrgb[0] = 0; matnum = 1;

		//Tri-mesh, Vertices: 16     Faces: 8
		//Vertex 0:  X: -128.239471     Y: -201...
		//Face 0:    A:0 B:1 C:2 AB:1 BC:1 CA:0
	if (!kzopen(fnam)) return(0);
	printf("Reading %s\n",fnam);
	hidden = 0;
	do
	{
			//Read 1 line and remove spaces at beginning...
		i = -1;
		while (1)
		{
			j = kzgetc(); if (j < 0) break;
			if ((j < 32) && (i >= 0)) { tempbuf[i] = 0; break; }
			if ((j > 32) && (i == -1)) i = 0;
			if ((i >= 0) && (i < sizeof(tempbuf)-1)) tempbuf[i++] = (char)j;
		}

		if ((i >= 8) && (!memcmp(tempbuf,"Tri-mesh",8)))
		{
			hidden = 0;
			trioff[numobjects] = numtris;
			vertoff[numobjects] = numverts;
			numobjects++;
			if (numobjects >= maxobjects)
			{
				maxobjects <<= 1;
#if (PRINTALLOC == 1)
				printf("maxobjects=%d (%dKB)\n",maxobjects,(maxobjects*8)>>10);
#endif
				if (!(trioff = (long *)realloc(trioff,sizeof(long)*maxobjects))) quitout("realloc failed: trioff");
				if (!(vertoff = (long *)realloc(vertoff,sizeof(long)*maxobjects))) quitout("realloc failed: vertoff");
			}
			continue;
		}
		if ((i >= 6) && (!memcmp(tempbuf,"Hidden",6))) { hidden = 1; continue; }
		if ((i >= 6) && (!memcmp(tempbuf,"Vertex",6)))
		{
			if ((tempbuf[7] < '0') || (tempbuf[7] > '9') || (hidden)) continue;
			op = 7; p = findchar(&tempbuf[op],':')+op;
			x = findchar(&tempbuf[p],'X')+p;
			y = findchar(&tempbuf[x],'Y')+x;
			z = findchar(&tempbuf[y],'Z')+y;
			tempbuf[x] = 0; tempbuf[y] = 0; tempbuf[z] = 0;
			j = atol(&tempbuf[op])+vertoff[numobjects-1];
			if ((unsigned long)j >= maxverts)
			{
				maxverts += (maxverts>>1);
#if (PRINTALLOC == 1)
				printf("maxverts=%d (%dKB)\n",maxverts,(maxverts*20)>>10);
#endif
				if (!(vert = (vertyp *)realloc(vert,sizeof(vertyp)*maxverts))) quitout("realloc failed: vert");
			}
			vert[j].x = (float)atof(&tempbuf[x+2]);
			vert[j].y = (float)atof(&tempbuf[y+2]);
			vert[j].z = (float)atof(&tempbuf[z+2]);
			if (j >= numverts) numverts = j+1;
			continue;
		}
		if ((i >= 4) && (!memcmp(tempbuf,"Face",4)))
		{
			if ((tempbuf[5] < '0') || (tempbuf[5] > '9') || (hidden)) continue;
			op = 5; p = findchar(&tempbuf[op],':')+op;
			x = findchar(&tempbuf[p],'A')+p;
			y = findchar(&tempbuf[x],'B')+x;
			z = findchar(&tempbuf[y],'C')+y;
			w = findchar(&tempbuf[z],'A')+z;
			tempbuf[x] = 0; tempbuf[y] = 0; tempbuf[z] = 0; tempbuf[w] = 0;
			j = atol(&tempbuf[op])+trioff[numobjects-1];
			if ((unsigned long)j >= maxtris)
			{
				maxtris += (maxtris>>1);
#if (PRINTALLOC == 1)
				printf("maxtris=%d (%dKB)\n",maxtris,(maxtris*16)>>10);
#endif
				if (!(tri = (trityp *)realloc(tri,sizeof(trityp)*maxtris))) quitout("realloc failed: tri");
			}
			tri[j].x = atol(&tempbuf[x+2])+vertoff[numobjects-1];
			tri[j].y = atol(&tempbuf[y+2])+vertoff[numobjects-1];
			tri[j].z = atol(&tempbuf[z+2])+vertoff[numobjects-1];
			lastface = j;
			if (j >= numtris) numtris = j+1;
			continue;
		}
		if ((i >= 8) && (!memcmp(tempbuf,"Material",8))) //Material:"PALM TREE TRUNK"
		{
			x = 8;
			y = findchar(&tempbuf[x],'\"')+x;
			z = findchar(&tempbuf[y+1],'\"')+y+1;
			y++; tempbuf[z] = 0; //&tempbuf[y] is now string in quotes (quotes stripped)

			for(j=matnum-1;j>0;j--)
				if (!strcmp(&nambuf[matstr[j]],&tempbuf[y]))
					{ tri[lastface].i = j; break; }
			if (!j) //Material string not in list, so add it...
			{
				j = strlen(&tempbuf[y]);
				if (matnum >= maxmat)
				{
					maxmat <<= 1;
#if (PRINTALLOC == 1)
					printf("maxmat=%d (%dKB)\n",maxmat,(maxmat*8)>>10);
#endif
					if (!(matstr = (long *)realloc(matstr,sizeof(long)*maxmat))) quitout("realloc failed: matstr");
					if (!(colrgb = (long *)realloc(colrgb,sizeof(long)*maxmat))) quitout("realloc failed: colrgb");
				}
				if (namplc+j >= maxnamsiz)
				{
					maxnamsiz = max(maxnamsiz<<1,namplc+j+1);
#if (PRINTALLOC == 1)
					printf("maxnamsiz=%d (%dKB)\n",maxnamsiz,(maxnamsiz*1)>>10);
#endif
					if (!(nambuf = (char *)realloc(nambuf,sizeof(char)*maxnamsiz))) quitout("realloc failed: nambuf");
				}
				tri[lastface].i = matnum;
				matstr[matnum] = namplc; matnum++;
				memcpy(&nambuf[namplc],&tempbuf[y],j+1); namplc += j+1;
			}
		}
	} while (!kzeof());
	kzclose();
	trioff[numobjects] = numtris;
	vertoff[numobjects] = numverts;

	if (matnum <= 2)
	{
		if (numobjects < 256)
		{
			for(i=0;i<numobjects;i++)
			{
				colrgb[i+1] = ((rand()&127)+64)+(((rand()&127)+64)<<8)+(((rand()&127)+64)<<16);
				for(z=trioff[i];z<trioff[i+1];z++) tri[z].i = i+1;
			}
			matnum = numobjects+1;
		}
		else
		{
			for(i=0;i<numtris;i++) tri[i].i = 1;
			colrgb[1] = 0x808080; matnum = 2;
		}
	}
	else
	{
		for(i=0;i<matnum;i++)
		{
			colrgb[i] = ((rand()&127)+64)+(((rand()&127)+64)<<8)+(((rand()&127)+64)<<16);
			x = 24; y = 0xff000000;
			for(j=0;nambuf[matstr[i]+j];j++)
			{
				switch(nambuf[matstr[i]+j])
				{
					case 'B': case 'b': x =  0; y = 0x000000ff; colrgb[i] &= ~y; break;
					case 'G': case 'g': x =  8; y = 0x0000ff00; colrgb[i] &= ~y; break;
					case 'R': case 'r': x = 16; y = 0x00ff0000; colrgb[i] &= ~y; break;
					case 'A': case 'a': x = 24; y = 0xff000000; colrgb[i] &= ~y; break;
					case '0': case '1': case '2': case '3': case '4':
					case '5': case '6': case '7': case '8': case '9':
						colrgb[i] = (((colrgb[i]&y)*10+((nambuf[matstr[i]+j]-48)<<x))&y)+(colrgb[i]&~y);
						break;
				}
			}
		}
	}

	return(1);
}

void chunkread(long p);

char *ASCIIZReader (long p)
{
	int i, j = 0;
	while ((i = kzgetc()) > 0)
		if (j < sizeof(tempbuf)-1) tempbuf[j++] = i;
	tempbuf[j] = 0;
	return(tempbuf);
}

void ObjBlockReader (long p)
{
	ASCIIZReader(p);
	chunkread(p); //Read rest of chunks inside this one.

	numobjects++;
	if (numobjects >= maxobjects)
	{
		maxobjects <<= 1;
#if (PRINTALLOC == 1)
		printf("maxobjects=%d (%dKB)\n",maxobjects,(maxobjects*8)>>10);
#endif
		if (!(trioff = (long *)realloc(trioff,sizeof(long)*maxobjects))) quitout("realloc failed: trioff");
		if (!(vertoff = (long *)realloc(vertoff,sizeof(long)*maxobjects))) quitout("realloc failed: vertoff");
	}
	trioff[numobjects] = numtris;
	vertoff[numobjects] = numverts;
}

void VertListReader (long p)
{
	unsigned short s;
	long nv;
	float c[3];

	if (!kzread(&s,sizeof(s))) return;
	nv = (long)s;
	//printf("vert %d\n",nv);
	while (nv-- > 0)
	{
		if (!kzread(&c,sizeof(c))) return;

		if ((unsigned long)numverts >= maxverts)
		{
			maxverts += (maxverts>>1);
#if (PRINTALLOC == 1)
			printf("maxverts=%d (%dKB)\n",maxverts,(maxverts*20)>>10);
#endif
			if (!(vert = (vertyp *)realloc(vert,sizeof(vertyp)*maxverts))) quitout("realloc failed: vert");
		}
		vert[numverts].x = c[0];
		vert[numverts].y = c[1];
		vert[numverts].z = c[2];
		numverts++;
	}
}

void FaceListReader (long p)
{
	unsigned short s, c[3], flg;
	long nv;

	if (!kzread(&s,sizeof(s))) return;
	nv = (long)s;
	//printf("face %d\n",nv);
	while (nv-- > 0)
	{
		if (!kzread(&c,sizeof(c))) return;
			//CA:flg&1, BC:flg&2, AB:flg&4, UWrap:flg&8, VWrap:flg&16
		if (!kzread(&flg,sizeof(flg))) return;

		if ((unsigned long)numtris >= maxtris)
		{
			maxtris += (maxtris>>1);
#if (PRINTALLOC == 1)
			printf("maxtris=%d (%dKB)\n",maxtris,(maxtris*16)>>10);
#endif
			if (!(tri = (trityp *)realloc(tri,sizeof(trityp)*maxtris))) quitout("realloc failed: tri");
		}
		tri[numtris].x = ((long)c[0])+vertoff[numobjects];
		tri[numtris].y = ((long)c[1])+vertoff[numobjects];
		tri[numtris].z = ((long)c[2])+vertoff[numobjects];
		numtris++;
	}
	chunkread(p); //Read rest of chunks inside this one.
}

long matstring2num (char *st)
{
	long j, k;

	for(j=matnum-1;j>0;j--)
		if (!strcmp(&nambuf[matstr[j]],st))
			break;
	if (!j) //Material string not in list, so add it...
	{
		k = strlen(st);
		if (matnum >= maxmat)
		{
			maxmat <<= 1;
#if (PRINTALLOC == 1)
			printf("maxmat=%d (%dKB)\n",maxmat,(maxmat*8)>>10);
#endif
			if (!(matstr = (long *)realloc(matstr,sizeof(long)*maxmat))) quitout("realloc failed: matstr");
			if (!(colrgb = (long *)realloc(colrgb,sizeof(long)*maxmat))) quitout("realloc failed: colrgb");
		}
		if (namplc+k >= maxnamsiz)
		{
			maxnamsiz = max(maxnamsiz<<1,namplc+k+1);
#if (PRINTALLOC == 1)
			printf("maxnamsiz=%d (%dKB)\n",maxnamsiz,(maxnamsiz*1)>>10);
#endif
			if (!(nambuf = (char *)realloc(nambuf,sizeof(char)*maxnamsiz))) quitout("realloc failed: nambuf");
		}
		j = matnum;
		matstr[matnum] = namplc;
		colrgb[matnum] = 0; matnum++;
		memcpy(&nambuf[namplc],st,k+1); namplc += k+1;
	}
	return(j);
}

void FaceMatReader (long p)
{
	unsigned short s, nf;
	long n, j;

	ASCIIZReader(p);
	//printf("%s\n",tempbuf);
	j = matstring2num(tempbuf);

	if (!kzread(&s,sizeof(s))) return;
	n = (long)s;
	while (n-- > 0)
	{
		if (!kzread(&nf,sizeof(nf))) return;
		//printf("tri[%d].i = %d\n",((long)nf)+trioff[numobjects],j);
		tri[((long)nf)+trioff[numobjects]].i = j;
	}
}

void MapListReader (long p)
{
	unsigned short s;
	long nv, j;
	float c[2];

	if (!kzread(&s,sizeof(s))) return;
	nv = (long)s; j = 0;
	while (nv-- > 0)
	{
		if (!kzread(&c,sizeof(c))) return;
		vert[vertoff[numobjects]+j].u = c[0];
		vert[vertoff[numobjects]+j].v = c[1];
		j++;
	}
}

static long lastmat = 0;
void MatNameReader (long p)
{
	ASCIIZReader(p);
	lastmat = matstring2num(tempbuf);
	//printf("%s\n",tempbuf);
}

long picstring2num (char *st)
{
	long j, k;

	for(j=numpics-1;j>=0;j--)
		if (!strcmp(&nambuf[pic[j].namoff],st))
			break;
	if (j < 0) //Material string not in list, so add it...
	{
		k = strlen(st);

		if (numpics >= maxpics)
		{
			maxpics <<= 1;
#if (PRINTALLOC == 1)
			printf("maxpics=%d, (%dKB)\n",maxpics,(maxpics*20)>>10);
#endif
			if (!(pic = (pictyp *)realloc(pic,sizeof(pictyp)*maxpics))) quitout("realloc failed: pic");
		}
		if (namplc+k >= maxnamsiz)
		{
			maxnamsiz = max(maxnamsiz<<1,namplc+k+1);
#if (PRINTALLOC == 1)
			printf("maxnamsiz=%d (%dKB)\n",maxnamsiz,(maxnamsiz*1)>>10);
#endif
			if (!(nambuf = (char *)realloc(nambuf,sizeof(char)*maxnamsiz))) quitout("realloc failed: nambuf");
		}
		j = numpics;
		pic[numpics].namoff = namplc;
		numpics++;
		memcpy(&nambuf[namplc],st,k+1); namplc += k+1;
	}
	return(j);
}

void MapFileReader (long p)
{
	ASCIIZReader(p);
	//printf("%s\n",tempbuf);
	colrgb[lastmat] = ~(picstring2num(tempbuf));
}

void RGBFReader (long p)
{
	float c[3];
	if (!kzread(&c,sizeof(c))) return;
	if (colrgb[lastmat] < 0) return; //use texture map instead!
	colrgb[lastmat] = (((long)c[0])<<16) + (((long)c[1])<<8) + ((long)c[2]);
	//printf("color=0x%08x\n",colrgb[lastmat]);

}

void RGBBReader (long p)
{
	char c[3];
	if (!kzread(&c,sizeof(c))) return;
	if (colrgb[lastmat] < 0) return; //use texture map instead!
	colrgb[lastmat] = (((long)c[0])<<16)+(((long)c[1])<<8)+((long)c[2]);
	//printf("color=0x%08x\n",colrgb[lastmat]);
}

struct
{
	unsigned short id;
	void (*func)(long p);
} chunknames[] =
{
	{0x0010,RGBFReader},     //RGB float
	{0x0011,RGBBReader},     //RGB char
	{0x4D4D,0},              //Main
	{0x3D3D,0},              //Object Mesh
	{0x4000,ObjBlockReader}, //Object Block
	{0x4100,0},              //Tri-Mesh
	{0x4110,VertListReader}, //Vertex list
	{0x4120,FaceListReader}, //Face list
	{0x4130,FaceMatReader},  //Face material
	{0x4140,MapListReader},  //Mappings list
	{0xAFFF,0},              //Material
	{0xA000,MatNameReader},  //Material name
	{0xA020,0},              //Diffuse color
	{0xA200,0},              //Texture map
	{0xA300,MapFileReader},  //Texture filename
};

void chunkread (long p)
{
	long n, pc, leng;
	unsigned short id;

	while (kztell() < p)
	{
		pc = kztell();
		if (!kzread(&id,2)) return;
		if (!kzread(&leng,4)) return; if (!leng) return;
		//printf("id=%04x, leng=%d\n",id,leng);
		for (n=(sizeof(chunknames)/sizeof(chunknames[0]))-1;n>=0;n--)
			if (chunknames[n].id == id)
			{
				pc += leng;
				if (chunknames[n].func) chunknames[n].func(pc);
										 else chunkread(pc);
				kzseek(pc,SEEK_SET);
				break;
			}
		if (n < 0) kzseek(pc+leng,SEEK_SET);
	}
}

long load3ds (char *fnam)
{
	long leng;

	numobjects = numverts = numtris = numpics = 0;

		//Make material index #0 a dummy color (black)
	matstr[0] = namplc;
	strcpy(&nambuf[namplc],"r0g0b0a0"); namplc += strlen(&nambuf[namplc])+1;
	matnum = 1;

	trioff[0] = 0;
	vertoff[0] = 0;

	if (!kzopen(fnam)) return(0);
	printf("Reading %s\n",fnam);
	kzseek(0,SEEK_END); leng = kztell(); kzseek(0,SEEK_SET);
	chunkread(leng);
	kzclose();

	return(1);
}

//-------------------------------------  MD2 LIBRARY BEGINS ----------------------------------------
	//This MD2 code is based on the source code from David Henry (tfc_duke(at)hotmail.com)
	//   http://tfc.duke.free.fr/us/tutorials/models/md2.htm
	//He probably wouldn't recognize it if he looked at it though :)


typedef struct
{  long id, vers, skinxsiz, skinysiz, framebytes; //id:"IPD2", vers:8
	long numskins, numverts, numuv, numtris, numglcmds, numframes;
	long ofsskins, ofsuv, ofstris, ofsframes, ofsglcmds, ofseof; //ofsskins: skin names (64 bytes each)
} md2head_t;

typedef struct { unsigned char v[3], ni; } md2vert_t; //compressed vertex coords (x,y,z)
typedef struct
{  point3d mul, add; //scale&translation vector
	char name[16];    //frame name
	md2vert_t verts[1]; //first vertex of this frame
} md2frame_t;

typedef struct
{
	long numframes, numverts, numglcmds, framebytes, *glcmds;
	float scale;
	char *frames;
} md2model;

static point3d *vertlist = 0;

void md2free (md2model *m)
{
	if (vertlist) { free(vertlist); vertlist = 0; }
	if (m->glcmds) { free(m->glcmds); m->glcmds = 0; }
	if (m->frames) { free(m->frames); m->frames = 0; }
}

long md2load (md2model *m, const char *filename)
{
	md2head_t head;
	char *buf, st[MAX_PATH+64];
	long i;

	memset(m,0,sizeof(md2model));
	m->glcmds = 0; m->numframes = m->numverts = m->numglcmds = 0; m->scale = 1; //.01;

	if (!kzopen(filename)) return(0);
	kzread((char *)&head,sizeof(md2head_t));
	if ((head.id != 0x32504449) && (head.vers != 8)) { kzclose(); return(0); } //"IDP2"

	m->numframes = head.numframes;
	m->numverts = head.numverts;
	m->numglcmds = head.numglcmds;
	m->framebytes = head.framebytes;
	m->frames = (char *)malloc(m->numframes*head.framebytes);
	m->glcmds = (long *)malloc(m->numglcmds*sizeof(long));
	kzseek(head.ofsframes,SEEK_SET); kzread((char *)m->frames,m->numframes*head.framebytes);
	kzseek(head.ofsglcmds,SEEK_SET); kzread((char *)m->glcmds,m->numglcmds*sizeof(long));

	strcpy(st,filename);
	for(i=strlen(st)-1;i>0;i--)
		if ((st[i] == '/') || (st[i] == '\\')) { i++; break; }
	kzseek(head.ofsskins,SEEK_SET); kzread(&st[i],64);
	for(;st[i];i++) if (st[i] == ':') st[i] = 0; //Hack for certain files
	kzclose();

	numpics = 1; i = strlen(st);
	if (namplc+i >= maxnamsiz)
	{
		maxnamsiz = max(maxnamsiz<<1,namplc+i+1);
#if (PRINTALLOC == 1)
		printf("maxnamsiz=%d (%dKB)\n",maxnamsiz,(maxnamsiz*1)>>10);
#endif
		if (!(nambuf = (char *)realloc(nambuf,sizeof(char)*maxnamsiz))) quitout("realloc failed: nambuf");
	}
	memcpy(&nambuf[namplc],st,i+1); namplc += i+1;
	colrgb[0] = -1;

	vertlist = (point3d *)malloc(m->numverts*sizeof(point3d));

	return(1);
}

void md2draw (md2model *m, long cframe, long nframe, float interp)
{
	point3d p0, p1, m0, m1, a0, a1;
	md2frame_t *f0, *f1;
	unsigned char *c0, *c1;
	long i, j, onumverts, trit, *lptr;
	float f, g;

		//create current&next frame's vertex list from whole list
	cframe = min(max(cframe,0),m->numframes-1);
	if ((unsigned long)nframe >= (unsigned long)m->numframes) nframe = cframe;
	f0 = (md2frame_t *)&m->frames[cframe*m->framebytes];
	f1 = (md2frame_t *)&m->frames[nframe*m->framebytes];
	f = interp; g = 1-f;
	m0.x = f0->mul.x*m->scale*g; m1.x = f1->mul.x*m->scale*f;
	m0.y = f0->mul.y*m->scale*g; m1.y = f1->mul.y*m->scale*f;
	m0.z = f0->mul.z*m->scale*g; m1.z = f1->mul.z*m->scale*f;
	a0.x = f0->add.x*m->scale; a0.x = (f1->add.x*m->scale-a0.x)*f+a0.x;
	a0.y = f0->add.y*m->scale; a0.y = (f1->add.y*m->scale-a0.y)*f+a0.y;
	a0.z = f0->add.z*m->scale; a0.z = (f1->add.z*m->scale-a0.z)*f+a0.z;
	c0 = &f0->verts[0].v[0]; c1 = &f1->verts[0].v[0];
	m0.x = -m0.x; m1.x = -m1.x; a0.x = -a0.x;

	f = 1*32.0;
	m0.x *= f; m0.y *= f; m0.z *= f;
	m1.x *= f; m1.y *= f; m1.z *= f;
	a0.x *= f; a0.y *= f; a0.z *= f;

	for(i=m->numverts-1;i>=0;i--) //interpolate&scale vertices to avoid ugly animation
	{
		vertlist[i].z = c0[(i<<2)+0]*m0.x + c1[(i<<2)+0]*m1.x + a0.x;
		vertlist[i].y = c0[(i<<2)+2]*m0.z + c1[(i<<2)+2]*m1.z + a0.z;
		vertlist[i].x = c0[(i<<2)+1]*m0.y + c1[(i<<2)+1]*m1.y + a0.y;
	}

	for(lptr=m->glcmds;i=(*lptr++);)
	{
		if (i < 0) { i = -i; trit = 0; /*TRI_FAN*/ } else { trit = 1; /*TRI_STRIP*/ }

		if ((unsigned long)(numverts+i) >= maxverts)
		{
			maxverts = max(maxverts+(maxverts>>1),numverts+i);
#if (PRINTALLOC == 1)
			printf("maxverts=%d (%dKB)\n",maxverts,(maxverts*20)>>10);
#endif
			if (!(vert = (vertyp *)realloc(vert,sizeof(vertyp)*maxverts))) quitout("realloc failed: vert");
		}
		j = i; onumverts = numverts;
		for(;i>0;i--,lptr+=3)
		{
			vert[numverts].x =-vertlist[lptr[2]].z;
			vert[numverts].y = vertlist[lptr[2]].x;
			vert[numverts].z = vertlist[lptr[2]].y;
			vert[numverts].u = (((float *)lptr)[0]);
			vert[numverts].v = 1-(((float *)lptr)[1]);
			numverts++;
		}

		if ((unsigned long)(numtris+j-2) >= maxtris)
		{
			maxtris = max(maxtris+(maxtris>>1),numtris+j-2);
#if (PRINTALLOC == 1)
			printf("maxtris=%d (%dKB)\n",maxtris,(maxtris*16)>>10);
#endif
			if (!(tri = (trityp *)realloc(tri,sizeof(trityp)*maxtris))) quitout("realloc failed: tri");
		}

		for(i=2;i<j;i++)
		{
			if (!trit)
			{
				tri[numtris].x = onumverts; //TRI_FAN
				tri[numtris].y = onumverts+i-1;
				tri[numtris].z = onumverts+i;
			}
			else
			{
				tri[numtris].x = onumverts+i-2+(i&1); //TRI_STRIP
				tri[numtris].y = onumverts+i-1-(i&1);
				tri[numtris].z = onumverts+i;
			}
			tri[numtris].i = 0;
			numtris++;
		}
	}
}

//--------------------------------------  MD2 LIBRARY ENDS -----------------------------------------
//-------------------------------------  MD3 LIBRARY BEGINS ----------------------------------------

typedef struct { char nam[64]; long i; } md3shader_t; //ascz path of shader, shader index
typedef struct { long i[3]; } md3tri_t; //indices of tri
typedef struct { float u, v; } md3uv_t;
typedef struct { signed short x, y, z; unsigned char nlat, nlng; } md3xyzn_t; //xyz are [10:6] ints

typedef struct
{
	point3d min, max, cen; //bounding box&origin
	float r; //radius of bounding sphere
	char nam[16]; //ascz frame name
} md3frame_t;

typedef struct
{
	char nam[64]; //ascz tag name
	point3d p, x, y, z; //tag object pos&orient
} md3tag_t;

typedef struct
{
	long id; //IDP3(0x33806873)
	char nam[64]; //ascz surface name
	long flags; //?
	long numframes, numshaders, numverts, numtris; //numframes same as md3head,max shade=~256,vert=~4096,tri=~8192
	md3tri_t *tris;       //file format: rel offs from md3surf
	md3shader_t *shaders; //file format: rel offs from md3surf
	md3uv_t *uv;          //file format: rel offs from md3surf
	md3xyzn_t *xyzn;      //file format: rel offs from md3surf
	long ofsend;
} md3surf_t;

typedef struct
{
	long id, vers; //id=IDP3(0x33806873), vers=15
	char nam[64]; //ascz path in PK3
	long flags; //?
	long numframes, numtags, numsurfs, numskins; //max=~1024,~16,~32,numskins=artifact of MD2; use shader field instead
	md3frame_t *frames; //file format: abs offs
	md3tag_t *tags;     //file format: abs offs
	md3surf_t *surfs;   //file format: abs offs
	long eof;           //file format: abs offs
} md3head_t;

typedef struct
{
	long mdnum; //MD2=2, MD3=3. NOTE: must be first in structure!
	md3head_t head;
	long frame0, frame1, cframe, nframe, fpssc, usesalpha;
	float oldtime, curtime, interpol, scale;
	unsigned int texid;
} md3model;

long md3load (md3model *m, const char *filnam)
{
	FILE *fil;
	char *buf, st[MAX_PATH+2], bst[MAX_PATH+2];
	long i, j, surfi, ofsurf, maxverts, bsc;
	md3surf_t *s;

	fil = fopen(filnam,"rb"); if (!fil) return(0);

	memset(m,0,sizeof(md3model)); m->mdnum = 3;
	m->texid = 0; m->scale = .01;

	fread(&m->head,sizeof(md3head_t),1,fil);
	if ((m->head.id != 0x33504449) && (m->head.vers != 15)) { fclose(fil); free(m); return(0); } //"IDP3"

	ofsurf = (long)m->head.surfs; maxverts = 0;

	fseek(fil,(long)m->head.frames,SEEK_SET); i = m->head.numframes*sizeof(md3frame_t);
	m->head.frames = (md3frame_t *)malloc(i); if (!m->head.frames) return(0);
	fread(m->head.frames,i,1,fil);

	fseek(fil,(long)m->head.tags,SEEK_SET); i = m->head.numtags*sizeof(md3tag_t);
	m->head.tags = (md3tag_t *)malloc(i); if (!m->head.tags) return(0);
	fread(m->head.tags,i,1,fil);

	fseek(fil,(long)m->head.surfs,SEEK_SET); i = m->head.numsurfs*sizeof(md3surf_t);
	m->head.surfs = (md3surf_t *)malloc(i); if (!m->head.surfs) return(0);
	for(surfi=0;surfi<m->head.numsurfs;surfi++)
	{
		s = &m->head.surfs[surfi];
		fseek(fil,ofsurf,SEEK_SET); fread(s,sizeof(md3surf_t),1,fil);

		fseek(fil,ofsurf+((long)(s->tris)),SEEK_SET); i = s->numtris*sizeof(md3tri_t);
		s->tris = (md3tri_t *)malloc(i); if (!s->tris) return(0);
		fread(s->tris,i,1,fil);

		fseek(fil,ofsurf+((long)(s->shaders)),SEEK_SET); i = s->numshaders*sizeof(md3shader_t);
		s->shaders = (md3shader_t *)malloc(i); if (!s->shaders) return(0);
		fread(s->shaders,i,1,fil);

		fseek(fil,ofsurf+((long)(s->uv)),SEEK_SET); i = s->numverts*sizeof(md3uv_t);
		s->uv = (md3uv_t *)malloc(i); if (!s->uv) return(0);
		fread(s->uv,i,1,fil);

		fseek(fil,ofsurf+((long)(s->xyzn)),SEEK_SET); i = s->numframes*s->numverts*sizeof(md3xyzn_t);
		s->xyzn = (md3xyzn_t *)malloc(i); if (!s->xyzn) return(0);
		fread(s->xyzn,i,1,fil);

		if (s->numverts > maxverts) maxverts = s->numverts;
		ofsurf += s->ofsend;
	}
	fclose(fil);

	m->frame0 = 0; m->frame1 = m->head.numframes;
	m->cframe = 0; m->nframe = m->cframe+1; if (m->nframe >= m->head.numframes) m->nframe = 0;
	m->fpssc = 8;

	strcpy(st,filnam);
	for(i=0,j=0;st[i];i++) if ((st[i] == '/') || (st[i] == '\\')) j = i+1;
	st[j] = '*'; st[j+1] = 0;

	kzfindfilestart(st); bsc = -1;
	while (kzfindfile(st))
	{
		if (st[0] == '\\') continue;

		for(i=0,j=0;st[i];i++) if (st[i] == '.') j = i+1;
		if ((!stricmp(&st[j],"JPG")) || (!stricmp(&st[j],"PNG")) || (!stricmp(&st[j],"GIF")) ||
			 (!stricmp(&st[j],"PCX")) || (!stricmp(&st[j],"TGA")) || (!stricmp(&st[j],"BMP")) ||
			 (!stricmp(&st[j],"CEL")))
		{
			for(i=0;st[i];i++) if (st[i] != filnam[i]) break;
			if (i > bsc) { bsc = i; strcpy(bst,st); }
		}
	}
	//if (!mdloadskin(&m->texid,&m->usesalpha,bst)) ;//bad!
	numpics = 1; i = strlen(bst);
	if (namplc+i >= maxnamsiz)
	{
		maxnamsiz = max(maxnamsiz<<1,namplc+i+1);
#if (PRINTALLOC == 1)
		printf("maxnamsiz=%d (%dKB)\n",maxnamsiz,(maxnamsiz*1)>>10);
#endif
		if (!(nambuf = (char *)realloc(nambuf,sizeof(char)*maxnamsiz))) quitout("realloc failed: nambuf");
	}
	memcpy(&nambuf[namplc],bst,i+1); namplc += i+1;
	colrgb[0] = -1;

	vertlist = (point3d *)malloc(maxverts*sizeof(point3d));

	return(1);
}

void md3draw (md3model *m, long cframe, long nframe, float interp)
{
	point3d p0, p1, m0, m1, a0, a1;
	md3frame_t *f0, *f1;
	md3xyzn_t *v0, *v1;
	long i, j, k, surfi, *lptr;
	float f, g;
	md3surf_t *s;

	cframe = min(max(cframe,0),m->head.numframes-1);
	if ((unsigned long)nframe >= (unsigned long)m->head.numframes) nframe = cframe;

		//create current&next frame's vertex list from whole list
	f0 = (md3frame_t *)&m->head.frames[cframe];
	f1 = (md3frame_t *)&m->head.frames[nframe];
	f = interp; g = 1-f;
	m0.x = (1.0/64.0)*m->scale*g; m1.x = (1.0/64.0)*m->scale*f;
	m0.y = (1.0/64.0)*m->scale*g; m1.y = (1.0/64.0)*m->scale*f;
	m0.z = (1.0/64.0)*m->scale*g; m1.z = (1.0/64.0)*m->scale*f;
	a0.x = f0->cen.x*m->scale; a0.x = (f1->cen.x*m->scale-a0.x)*f+a0.x;
	a0.y = f0->cen.y*m->scale; a0.y = (f1->cen.y*m->scale-a0.y)*f+a0.y;
	a0.z = f0->cen.z*m->scale; a0.z = (f1->cen.z*m->scale-a0.z)*f+a0.z;
	m0.x = -m0.x; m1.x = -m1.x; a0.x = -a0.x;

	f = 1*32.0;
	m0.x *= f; m0.y *= f; m0.z *= f;
	m1.x *= f; m1.y *= f; m1.z *= f;
	a0.x *= f; a0.y *= f; a0.z *= f;

	for(surfi=0;surfi<m->head.numsurfs;surfi++)
	{
		s = &m->head.surfs[surfi];
		v0 = &s->xyzn[cframe*s->numverts];
		v1 = &s->xyzn[nframe*s->numverts];
		for(i=s->numverts-1;i>=0;i--) //interpolate&scale vertices to avoid ugly animation
		{
			vertlist[i].z = ((float)v0[i].x)*m0.x + ((float)v1[i].x)*m1.x + a0.x;
			vertlist[i].y = ((float)v0[i].z)*m0.z + ((float)v1[i].z)*m1.z + a0.z;
			vertlist[i].x = ((float)v0[i].y)*m0.y + ((float)v1[i].y)*m1.y + a0.y;
		}

		i = s->numtris*3;
		if ((unsigned long)(numverts+i) >= maxverts)
		{
			maxverts = max(maxverts+(maxverts>>1),numverts+i);
#if (PRINTALLOC == 1)
			printf("maxverts=%d (%dKB)\n",maxverts,(maxverts*20)>>10);
#endif
			if (!(vert = (vertyp *)realloc(vert,sizeof(vertyp)*maxverts))) quitout("realloc failed: vert");
		}

		i = s->numtris;
		if ((unsigned long)(numtris+i) >= maxtris)
		{
			maxtris = max(maxtris+(maxtris>>1),numtris+i);
#if (PRINTALLOC == 1)
			printf("maxtris=%d (%dKB)\n",maxtris,(maxtris*16)>>10);
#endif
			if (!(tri = (trityp *)realloc(tri,sizeof(trityp)*maxtris))) quitout("realloc failed: tri");
		}

		//if (m->usesalpha) ???

		for(i=s->numtris-1;i>=0;i--)
		{
			tri[numtris].x = numverts;
			tri[numtris].y = numverts+1;
			tri[numtris].z = numverts+2;
			tri[numtris].i = 0;
			numtris++;

			for(j=0;j<3;j++)
			{
				k = s->tris[i].i[j];
				vert[numverts].x =-vertlist[k].z;
				vert[numverts].y = vertlist[k].x;
				vert[numverts].z = vertlist[k].y;
				vert[numverts].u = s->uv[k].u;
				vert[numverts].v = 1-s->uv[k].v;
				numverts++;
			}
		}
	}

#if 0
		//precalc:
	float sinlut256[256+(256>>2)];
	for(i=0;i<sizeof(sinlut256)/sizeof(sinlut256[0]);i++) sinlut256[i] = sin(((float)i)*(PI*2/255.0));

		//normal to xyz:
	md3vert_t *mv = &md3vert[?];
	z = sinlut256[mv->nlng+(256>>2)];
	x = sinlut256[mv->nlat]*z;
	y = sinlut256[mv->nlat+(256>>2)]*z;
	z = sinlut256[mv->nlng];
#endif
}

void md3free (md3model *m)
{
	md3surf_t *s;
	long surfi;

	//if (gtexmalloc) { free((void *)gtexmalloc); gtexmalloc = 0; }
	if (vertlist) { free(vertlist); vertlist = 0; }
	if (m)
	{
		//glDeleteLists(m->texid,1);
		if (m->head.surfs)
		{
			for(surfi=m->head.numsurfs-1;surfi>=0;surfi--)
			{
				s = &m->head.surfs[surfi];
				if (s->xyzn) free(s->xyzn);
				if (s->uv) free(s->uv);
				if (s->shaders) free(s->shaders);
				if (s->tris) free(s->tris);
			}
			free(m->head.surfs);
		}
		if (m->head.tags) free(m->head.tags);
		if (m->head.frames) free(m->head.frames);
		free(m);
	}
}

//--------------------------------------  MD3 LIBRARY ENDS -----------------------------------------

long gcframe = 0, gnframe = -1;
float ginterp = 0.f;
long loadmd2 (char *fnam)
{
	md2model mymodel;

	numverts = numtris = numpics = 0;

	printf("Reading %s\n",fnam);
	if (!md2load(&mymodel,fnam)) return(0);
	md2draw(&mymodel,gcframe,gnframe,ginterp);
	md2free(&mymodel);
	return(1);
}

long loadmd3 (char *fnam)
{
	md3model mymodel;

	numverts = numtris = numpics = 0;

	printf("Reading %s\n",fnam);
	if (!md3load(&mymodel,fnam)) return(0);
	md3draw(&mymodel,gcframe,gnframe,ginterp);
	md3free(&mymodel);
	return(1);
}

void hollowfix (long xsiz, long ysiz, long zsiz)
{
	long i, x, y, z;

		//Set all seeable 0's to 1's
	if (savetype != 3)
	{
		for(x=0;x<xsiz;x++)
			for(y=0;y<ysiz;y++)
				{ floodfill3dbits(x,y,0L); floodfill3dbits(x,y,zsiz-1); }
		for(x=0;x<xsiz;x++)
			for(z=0;z<zsiz;z++)
				{ floodfill3dbits(x,0L,z); floodfill3dbits(x,ysiz-1,z); }
		for(y=0;y<ysiz;y++)
			for(z=0;z<zsiz;z++)
				{ floodfill3dbits(0L,y,z); floodfill3dbits(xsiz-1,y,z); }
	}
	else
	{
		for(x=0;x<xsiz;x++)
			for(y=0;y<ysiz;y++)
				floodfill3dbits(x,y,0L);
	}

		//Invert all solid & air
	for(z=((xsiz*MAYDIM*MAZDIM)>>5)-1;z>=0;z--) vbit[z] ^= -1;

		//Make surface voxels solid again
	for(z=hashcnt-1;z>=0;z--) { x = hashdat[z].y; vbit[x>>5] |= (1<<x); }
}

void getboundaries ()
{
	long x, y, z;

	while (xsiz > 0)
	{
		for(y=ysiz-1;y>=0;y--)
			for(z=zsiz-1;z>=0;z--)
				if (vbit[(((xsiz-1)*MAYDIM+y)*MAZDIM+z)>>5]&(1<<z)) goto endxopt;
		xsiz--;
	}
endxopt:;
	while (ysiz > 0)
	{
		for(x=xsiz-1;x>=0;x--)
			for(z=zsiz-1;z>=0;z--)
				if (vbit[((x*MAYDIM+(ysiz-1))*MAZDIM+z)>>5]&(1<<z)) goto endyopt;
		ysiz--;
	}
endyopt:;
	while (zsiz > 0)
	{
		for(x=xsiz-1;x>=0;x--)
			for(y=ysiz-1;y>=0;y--)
				if (vbit[((x*MAYDIM+y)*MAZDIM+(zsiz-1))>>5]&(1<<(zsiz-1))) goto endzopt;
		zsiz--;
	}
endzopt:;
}

void getpal ()
{
	FILE *fil;
	long i, j, k, p, x, y, z;

	if (picsloaded > 0)
	{
		fil = 0;
		if (palfilnam[0]) fil = fopen(palfilnam,"rb");
		if (fil)
		{
			for(i=0;i<256;i++)
			{
				rpal[i] = (fgetc(fil)<<2);
				gpal[i] = (fgetc(fil)<<2);
				bpal[i] = (fgetc(fil)<<2);
			}
			fclose(fil);

			j = 0;
			for(i=64;i>=0;i--)
			{
				//j = (i-64)*(i-64);
				rdist[i] = rdist[128-i] = j*30;
				gdist[i] = gdist[128-i] = j*59;
				bdist[i] = bdist[128-i] = j*11;
				j += 129-(i<<1);
			}

			memset(closestcol,-1,sizeof(closestcol));
		}
		else
		{
				//Generate frequency distribution
			memset(palfreq,0,sizeof(palfreq));
			for(x=0;x<xsiz;x++)
				for(y=0;y<ysiz;y++)
				{
					p = (x*MAYDIM+y)*MAZDIM; z = dntil1(&vbit[p>>5],0);
					for(;z < zsiz;z=dntil1(&vbit[p>>5],z+1))
					{
						i = p+z;
						if (!(vbit[i>>5]&(1<<i))) continue;

						j = 0;
						if ((x == 0)      || (!(vbit[(i-MAYDIM*MAZDIM)>>5]&(1<<(i-MAYDIM*MAZDIM))))) j |= 1;
						if ((x == xsiz-1) || (!(vbit[(i+MAYDIM*MAZDIM)>>5]&(1<<(i+MAYDIM*MAZDIM))))) j |= 2;
						if ((y == 0)      || (!(vbit[(i-MAZDIM       )>>5]&(1<<(i-MAZDIM       ))))) j |= 4;
						if ((y == ysiz-1) || (!(vbit[(i+MAZDIM       )>>5]&(1<<(i+MAZDIM       ))))) j |= 8;
						if ((z == 0)      || (!(vbit[(i-1            )>>5]&(1<<(i-1            ))))) j |= 16;
						if ((z == zsiz-1) || (!(vbit[(i+1            )>>5]&(1<<(i+1            ))))) j |= 32;
						if (!j) continue;

						k = getpix32(x,y,z);
						palfreq[0][0][((k>>6)&0x3f000)+((k>>4)&0xfc0)+((k>>2)&0x3f)]++;
					}
				}

			genpal();
		}
	}
	else if (matnum <= 1)
	{
		colrgb[0] = 192+32;
	}
}

void writepal (FILE *fil)
{
	long i, j, k, x, y, z;

	if (picsloaded > 0)
	{
			//Store close-fit palette (object has textures)
		for(i=0;i<256;i++)
		{
			fputc(rpal[i]>>2,fil);
			fputc(gpal[i]>>2,fil);
			fputc(bpal[i]>>2,fil);
		}
	}
	else
	{
			//Generate interesting color palette
		if (matnum <= 1)
		{
			for(z=0;z<64;z++) { fputc(z,fil); fputc(0,fil); fputc(0,fil); }
			for(z=0;z<64;z++) { fputc(0,fil); fputc(z,fil); fputc(0,fil); }
			for(z=0;z<64;z++) { fputc(0,fil); fputc(0,fil); fputc(z,fil); }
			for(z=0;z<64;z++) { fputc(z,fil); fputc(z,fil); fputc(z,fil); }
		}
		else
		{
			for(i=0;i<matnum;i++)
			{
				fputc((colrgb[i]>>18)&63,fil);
				fputc((colrgb[i]>>10)&63,fil);
				fputc((colrgb[i]>> 2)&63,fil);
			}
				//Generate darker colors for palette
			k = ((256*2+i*1)/3);
			j = 1; y = (k-matnum)/(matnum-1); x = y-1;
			for(;i<k;i++)
			{
				if (!y) { fputc(0,fil); fputc(0,fil); fputc(0,fil); continue; }
				fputc(max((((colrgb[j]>>16)&255)*x)/(y<<2),0),fil);
				fputc(max((((colrgb[j]>> 8)&255)*x)/(y<<2),0),fil);
				fputc(max((((colrgb[j]    )&255)*x)/(y<<2),0),fil);
				j++; if (j >= matnum) { j = 1; x--; }
			}
				//Generate brighter colors for palette
			y = (256-i)/(matnum-1); x = y+1;
			for(;i<256;i++)
			{
				if (!y) { fputc(0,fil); fputc(0,fil); fputc(0,fil); continue; }
				fputc(min((((colrgb[j]>>16)&255)*x)/(y<<2),63),fil);
				fputc(min((((colrgb[j]>> 8)&255)*x)/(y<<2),63),fil);
				fputc(min((((colrgb[j]    )&255)*x)/(y<<2),63),fil);
				j++; if (j >= matnum) { j = 1; x++; }
			}
		}
	}
}

void savevox (char *fnam)
{
	FILE *fil;
	long i, j, k, p, x, y, z;

	getboundaries();
	if (!rendermode) hollowfix(xsiz,ysiz,zsiz);
	printf("Writing %s (%dx%dx%d)\n",fnam,xsiz,ysiz,zsiz);

	getpal();

	if (!(fil = fopen(fnam,"wb"))) quitout("Could not write file");
	fwrite(&xsiz,4,1,fil);
	fwrite(&ysiz,4,1,fil);
	fwrite(&zsiz,4,1,fil);

	for(x=0;x<xsiz;x++)
		for(y=0;y<ysiz;y++)
			for(z=0;z<zsiz;z++)
			{
				i = (x*MAYDIM+y)*MAZDIM+z;
				if (!(vbit[i>>5]&(1<<i))) { fputc(255,fil); continue; }

				j = 0;
				if ((x == 0)      || (!(vbit[(i-MAYDIM*MAZDIM)>>5]&(1<<(i-MAYDIM*MAZDIM))))) j |= 1;
				if ((x == xsiz-1) || (!(vbit[(i+MAYDIM*MAZDIM)>>5]&(1<<(i+MAYDIM*MAZDIM))))) j |= 2;
				if ((y == 0)      || (!(vbit[(i-MAZDIM       )>>5]&(1<<(i-MAZDIM       ))))) j |= 4;
				if ((y == ysiz-1) || (!(vbit[(i+MAZDIM       )>>5]&(1<<(i+MAZDIM       ))))) j |= 8;
				if ((z == 0)      || (!(vbit[(i-1            )>>5]&(1<<(i-1            ))))) j |= 16;
				if ((z == zsiz-1) || (!(vbit[(i+1            )>>5]&(1<<(i+1            ))))) j |= 32;
				if (!j) { fputc(0,fil); continue; }

				fputc(getclosestcol(getpix32(x,y,z),255),fil);
			}

	writepal(fil);

	fclose(fil);
}

void savekvx (char *fnam, long numips)
{
	FILE *fil;
	long i, j, k, l, x, y, z, o, p, m, oneupm, seekpos;
	long nxsiz, nysiz, nzsiz, vis, r, g, b, n, xx, yy, zz, xxx, yyy, zzz;
	char voxdat[MAZDIM+4];

	getboundaries();
	if (!rendermode) hollowfix(xsiz,ysiz,zsiz);
	printf("Writing %s (%dx%dx%d)\n",fnam,xsiz,ysiz,zsiz);

	getpal();

	nxsiz = xsiz; nysiz = ysiz; nzsiz = zsiz;

	if (!(fil = fopen(fnam,"wb"))) quitout("Could not write file");
	for(m=0;m<numips;m++)
	{
		oneupm = (1<<m);

		seekpos = ftell(fil);
		i = 0; fwrite(&i,4,1,fil);
		fwrite(&nxsiz,4,1,fil);
		fwrite(&nysiz,4,1,fil);
		fwrite(&nzsiz,4,1,fil);
		i = (((long)(xpiv*256.0))>>m); fwrite(&i,4,1,fil);
		i = (((long)(ypiv*256.0))>>m); fwrite(&i,4,1,fil);
		i = (((long)(zpiv*256.0))>>m); fwrite(&i,4,1,fil);
		i = ((nxsiz+1)<<2) + (((nysiz+1)*nxsiz)<<1); //xoffset+xyoffset table sizes
		fseek(fil,i,SEEK_CUR);

		for(x=0;x<nxsiz;x++)
		{
			xstart[x] = i;
			for(y=0;y<nysiz;y++)
			{
				xyoffs[x][y] = (short)(i-xstart[x]);
				j = 0;
				k = ((x*MAYDIM)+y)*MAZDIM; xxx = (x<<m); yyy = (y<<m);
				for(z=0;z<nzsiz;z++)
				{
					l = (1<<z); if (!(vbit[(k+z)>>5]&l)) continue;

						//Get vis
					vis = 0;
					if ((x ==       0) || (!(vbit[(k+z-MAYDIM*MAZDIM)>>5]&l))) vis |= 1;
					if ((x == nxsiz-1) || (!(vbit[(k+z+MAYDIM*MAZDIM)>>5]&l))) vis |= 2;
					if ((y ==       0) || (!(vbit[(k+z       -MAZDIM)>>5]&l))) vis |= 4;
					if ((y == nysiz-1) || (!(vbit[(k+z       +MAZDIM)>>5]&l))) vis |= 8;
					if ((z ==       0) || (!(vbit[(k+z  -1)>>5]&_lrotr(l,1)))) vis |= 16;
					if ((z == nzsiz-1) || (!(vbit[(k+z  +1)>>5]&_lrotl(l,1)))) vis |= 32;
					if (!vis) continue;

						//Get col
					r = g = b = n = 0; zzz = (z<<m);
					for(xx=min(xxx+oneupm,xsiz)-1;xx>=xxx;xx--)
						for(yy=min(yyy+oneupm,ysiz)-1;yy>=yyy;yy--)
							for(zz=min(zzz+oneupm,zsiz)-1;zz>=zzz;zz--)
							{
								if (getval((xx*MAYDIM+yy)*MAZDIM+zz) < 0) continue;
								o = getpix32(xx,yy,zz);
								r += ((o>>16)&255);
								g += ((o>> 8)&255);
								b += ( o     &255);
								n++;
							}
					if (n)
					{
						r = (r+(n>>1))/n;
						g = (g+(n>>1))/n;
						b = (b+(n>>1))/n;
					}

					if ((!j) || (z > voxdat[0]+voxdat[1]))
					{
						if (j) { fwrite(voxdat,j,1,fil); i += j; }
						voxdat[0] = z; voxdat[1] = voxdat[2] = 0; j = 3;
					}
					voxdat[1]++; voxdat[2] |= vis; voxdat[j++] = getclosestcol((r<<16)+(g<<8)+b,256);
				}
				if (j) { fwrite(voxdat,j,1,fil); i += j; }
			}
			xyoffs[x][y] = (short)(i-xstart[x]);
		}
		xstart[x] = i;

		j = ftell(fil)-seekpos-4;
		fseek(fil,seekpos,SEEK_SET);
		fwrite(&j,4,1,fil);
		fseek(fil,seekpos+28,SEEK_SET);
		fwrite(xstart,(nxsiz+1)<<2,1,fil);
		for(i=0;i<nxsiz;i++) fwrite(&xyoffs[i][0],(nysiz+1)<<1,1,fil);
		fseek(fil,seekpos+j+4,SEEK_SET);

		if (m < numips-1) //Halve vbit in all dimensions (which corrupts it)
		{
			if (nxsiz&1)
			{
				for(y=0;y<nysiz;y++)
					for(z=0;z<nzsiz;z++)
					{
						x = nxsiz-1;
						if (vbit[((x*MAYDIM+y)*MAZDIM+z)>>5]&(1<<z)) vbit[((nxsiz*MAYDIM+y)*MAZDIM+z)>>5] |= (1<<z);
																			 else vbit[((nxsiz*MAYDIM+y)*MAZDIM+z)>>5] &=~(1<<z);
					}
				nxsiz++;
			}
			if (nysiz&1)
			{
				for(x=0;x<nxsiz;x++)
					for(z=0;z<nzsiz;z++)
					{
						y = nysiz-1;
						if (vbit[((x*MAYDIM+y)*MAZDIM+z)>>5]&(1<<z)) vbit[((x*MAYDIM+nysiz)*MAZDIM+z)>>5] |= (1<<z);
																			 else vbit[((x*MAYDIM+nysiz)*MAZDIM+z)>>5] &=~(1<<z);
					}
				nysiz++;
			}
			if (nzsiz&1)
			{
				for(x=0;x<nxsiz;x++)
					for(y=0;y<nysiz;y++)
					{
						z = nzsiz-1;
						if (vbit[((x*MAYDIM+y)*MAZDIM+z)>>5]&(1<<z)) vbit[((x*MAYDIM+y)*MAZDIM+nzsiz)>>5] |= (1<<nzsiz);
																			 else vbit[((x*MAYDIM+y)*MAZDIM+nzsiz)>>5] &=~(1<<nzsiz);
					}
				nzsiz++;
			}

			nxsiz >>= 1; nysiz >>= 1; nzsiz >>= 1;
			for(x=0;x<nxsiz;x++)
				for(y=0;y<nysiz;y++)
				{
					l = (x*MAYDIM+y)*MAZDIM; k = (l<<1);
					for(z=0;z<nzsiz;z++)
					{
						j = 0; //# air voxels
						if (vbit[(k+(z<<1)                     )>>5]&(1<<(z<<1))) j++;
						if (vbit[(k+(z<<1)                     )>>5]&(2<<(z<<1))) j++;
						if (vbit[(k+(z<<1)+MAZDIM              )>>5]&(1<<(z<<1))) j++;
						if (vbit[(k+(z<<1)+MAZDIM              )>>5]&(2<<(z<<1))) j++;
						if (vbit[(k+(z<<1)       +MAYDIM*MAZDIM)>>5]&(1<<(z<<1))) j++;
						if (vbit[(k+(z<<1)       +MAYDIM*MAZDIM)>>5]&(2<<(z<<1))) j++;
						if (vbit[(k+(z<<1)+MAZDIM+MAYDIM*MAZDIM)>>5]&(1<<(z<<1))) j++;
						if (vbit[(k+(z<<1)+MAZDIM+MAYDIM*MAZDIM)>>5]&(2<<(z<<1))) j++;
						if (j >= 4) vbit[(l+z)>>5] |= (1<<z); //NOTE: vbit is inversed from SLAB6.C; Use >= 4
								 else vbit[(l+z)>>5] &=~(1<<z);
					}
				}
		}
	}

	writepal(fil);
	fclose(fil);
}

void savekv6 (char *fnam)
{
	FILE *fil;
	float f;
	long i, j, k, x, y, z, numvoxs;
	unsigned short u;

	getboundaries();
	if (!rendermode) hollowfix(xsiz,ysiz,zsiz);
	printf("Writing %s (%dx%dx%d)\n",fnam,xsiz,ysiz,zsiz);

	if (!(fil = fopen(fnam,"wb"))) quitout("Could not write file");

	i = 0x6c78764b;
	fwrite(&i,4,1,fil);
	fwrite(&xsiz,4,1,fil);
	fwrite(&ysiz,4,1,fil);
	fwrite(&zsiz,4,1,fil);
	f = (float)xpiv; fwrite(&f,4,1,fil); //Copy to f just in case I change *piv to double someday...
	f = (float)ypiv; fwrite(&f,4,1,fil);
	f = (float)zpiv; fwrite(&f,4,1,fil);

	equivecinit(255);

	numvoxs = 0;
	fwrite(&numvoxs,4,1,fil); //to be rewritten later

	for(x=0;x<xsiz;x++)
	{
		xstart[x] = 0;
		for(y=0;y<ysiz;y++)
		{
			xyoffs[x][y] = 0;
			for(z=0;z<zsiz;z++)
			{
				i = (x*MAYDIM+y)*MAZDIM+z;
				if (!(vbit[i>>5]&(1<<i))) continue;

				j = 0;
				if ((x == 0)      || (!(vbit[(i-MAYDIM*MAZDIM)>>5]&(1<<(i-MAYDIM*MAZDIM))))) j |= 1;
				if ((x == xsiz-1) || (!(vbit[(i+MAYDIM*MAZDIM)>>5]&(1<<(i+MAYDIM*MAZDIM))))) j |= 2;
				if ((y == 0)      || (!(vbit[(i-MAZDIM       )>>5]&(1<<(i-MAZDIM       ))))) j |= 4;
				if ((y == ysiz-1) || (!(vbit[(i+MAZDIM       )>>5]&(1<<(i+MAZDIM       ))))) j |= 8;
				if ((z == 0)      || (!(vbit[(i-1            )>>5]&(1<<(i-1            ))))) j |= 16;
				if ((z == zsiz-1) || (!(vbit[(i+1            )>>5]&(1<<(i+1            ))))) j |= 32;
				if (!j) continue;

				k = getpix32(x,y,z);
				fputc((k&255),fil);
				fputc((k>>8)&255,fil);
				fputc((k>>16)&255,fil);
				fputc(128,fil);
				fputc(z&255,fil); fputc(z>>8,fil);
				fputc(j,fil); //vis (0-63)
				fputc(getnorm8(x,y,z),fil); //dir (0-255)

				xyoffs[x][y]++; xstart[x]++; numvoxs++;
			}
		}
	}

	for(x=0;x<xsiz;x++) { i = xstart[x]; fwrite(&i,4,1,fil); }
	for(x=0;x<xsiz;x++)
		for(y=0;y<ysiz;y++) { u = xyoffs[x][y]; fwrite(&u,2,1,fil); }

	j = ftell(fil);
	fseek(fil,28,SEEK_SET);
	fwrite(&numvoxs,4,1,fil);
	fseek(fil,j,SEEK_SET);
	fclose(fil);
}

	//Returns !=0 if any voxels in (x,y,z0) to (x,y,z1-1) are solid, else 0
long anyvoxelsolid (long x, long y, long z0, long z1)
{
	if ((unsigned long)(x|y) >= VSID) return(0);
	if (z0 < 0) z0 = 0;
	if (z1 <= z0) return(0);
	if (z1 > MAZDIM) return(1);

	x = (x*MAYDIM + y)*MAZDIM;
	z0 += x; z1 += x;
	if (!((z0^z1)&~31)) return(vbit[z0>>5]&(~(-1<<z1))&(-1<<z0));
	x = (z0>>5);
	if (vbit[x]&(-1<<z0)) return(1);
	for(x++;x<(z1>>5);x++) if (vbit[x]) return(1);
	return(vbit[x]&(~(-1<<z1)));
}

	//Returns 0 if any voxel inside (vx-x)+(vy-y)+(vz-z) < r is solid
	//   NOTE: code uses really cool & optimized algorithm!
long issphereempty (long x, long y, long z, long r)
{
	long xx, yy, zz, k, r2, nr, z2;

	if (r <= 0) return(1);
	if (anyvoxelsolid(x,y,z-r+1,z+r)) return(0);
	r2 = r*r; nr = r2;
	for(yy=1;yy<r;yy++)
	{
		nr += 1-(yy<<1); k = nr; zz = r; z2 = r2;
		while (z2 >= k) { z2 += 1-(zz<<1); zz--; }
		if (anyvoxelsolid(x-yy,y,z-zz,z+zz+1)) return(0);
		if (anyvoxelsolid(x+yy,y,z-zz,z+zz+1)) return(0);
		if (anyvoxelsolid(x,y-yy,z-zz,z+zz+1)) return(0);
		if (anyvoxelsolid(x,y+yy,z-zz,z+zz+1)) return(0);
		for(xx=1;xx<yy;xx++)
		{
			k += 1-(xx<<1); if (k <= 0) break;
			while (z2 >= k) { z2 += 1-(zz<<1); zz--; }
			if (zz < 0) break;
			if (anyvoxelsolid(x-xx,y-yy,z-zz,z+zz+1)) return(0);
			if (anyvoxelsolid(x+xx,y-yy,z-zz,z+zz+1)) return(0);
			if (anyvoxelsolid(x-yy,y-xx,z-zz,z+zz+1)) return(0);
			if (anyvoxelsolid(x+yy,y-xx,z-zz,z+zz+1)) return(0);
			if (anyvoxelsolid(x-yy,y+xx,z-zz,z+zz+1)) return(0);
			if (anyvoxelsolid(x+yy,y+xx,z-zz,z+zz+1)) return(0);
			if (anyvoxelsolid(x-xx,y+yy,z-zz,z+zz+1)) return(0);
			if (anyvoxelsolid(x+xx,y+yy,z-zz,z+zz+1)) return(0);
		}
		k += 1-(xx<<1);
		if (k > 0)
		{
			while (z2 >= k) { z2 += 1-(zz<<1); zz--; }
			if (anyvoxelsolid(x-yy,y-yy,z-zz,z+zz+1)) return(0);
			if (anyvoxelsolid(x-yy,y+yy,z-zz,z+zz+1)) return(0);
			if (anyvoxelsolid(x+yy,y-yy,z-zz,z+zz+1)) return(0);
			if (anyvoxelsolid(x+yy,y+yy,z-zz,z+zz+1)) return(0);
		}
	}
	return(1);
}

#if 0

	//WARNING: OBSOLETE CODE - USE THE OTHER COMPILEBITSTACK() FROM THIS SOURCE!
	//NOTE: copied from VOXLAP5.C but with modified in these ways:
	//   1.change code near "colfunc"
	//   2.converted longs to bits
	//   3.pre-calculate 4 neighbors (n0,n1,n2,n3) & pass 1 neighbof buf (n0)
	// Inputs: uind[MAXZDIM]: uncompressed bitbuf (0=air,1=solid)
	//         n0[MAXZDIM]: neighbor bitbuf (n0&n1&n2&n3): (0=air,1=solid)
	//         px,py: parameters for setting unexposed voxel colors
	//Outputs: cbuf[MAXCSIZ]: compressed output buffer
	//Returns: n: length of compressed buffer (in bytes)
long compilebitstack (long *uind, long *n0, char *cbuf, long px, long py)
{
	long oz, onext, n, cp2, cp1, cp0, rp1, rp0, np, ns;
	lpoint3d p;

	p.x = px; p.y = py;

		//Do top slab (sky)
	oz = -1;
	p.z = -1; while (!(uind[(p.z+1)>>5]&(1<<(p.z+1)))) p.z++;
	onext = 0;
	cbuf[1] = p.z+1;
	cbuf[2] = p.z+1;
	cbuf[3] = 0;  //Top z0 (filler, not used yet)
	n = 4;
	cp1 = 1; cp0 = 0;
	rp1 = -1; rp0 = -1;

	do
	{
			//cp2 = state at p.z-1 (0 = air, 1 = next2air, 2 = solid)
			//cp1 = state at p.z   (0 = air, 1 = next2air, 2 = solid)
			//cp0 = state at p.z+1 (0 = air, 1 = next2air, 2 = solid)
		cp2 = cp1; cp1 = cp0; cp0 = 2;
		if (p.z < MAXZDIM-2)  //Bottom must be solid!
		{
			ns = ((p.z+1)>>5); np = (1<<(p.z+1));
				  if (!(uind[ns]&np)) cp0 = 0;
			else if (!(  n0[ns]&np)) cp0 = 1;
		}

			//Add slab
		if (cp1 != rp0)
		{
			if ((!cp1) && (rp0 > 0)) { oz = p.z; }
			else if ((rp0 < cp1) && (rp0 < rp1))
			{
				if (oz < 0) oz = p.z;
				cbuf[onext] = ((n-onext)>>2); onext = n;
				cbuf[n+1] = p.z;
				cbuf[n+2] = p.z-1;
				cbuf[n+3] = oz;
				n += 4; oz = -1;
			}
			rp1 = rp0; rp0 = cp1;
		}

			//Add color
		if ((cp1 == 1) || ((cp1 == 2) && ((!cp0) || (!cp2))))
		{
			if (cbuf[onext+2] == p.z-1) cbuf[onext+2] = p.z;
			*(long *)&cbuf[n] = ((getpix32(p.x,p.y,p.z)&0xffffff)|0x80000000);
			n += 4;
		}

		p.z++;
	} while (p.z < MAXZDIM);
	cbuf[onext] = 0;
	return(n);
}

#else

	//07/09/2004: See RAYRECU4.BAS (beginning of compilestack) for derivation.
	//   Code is equivalent to the original compilebitstack() converted from
	//VOXLAP5.C's compilestack (and first used in POLY2VOX). The main difference
	//is this version is much faster because it uses dntil0()/dntil1()! :) The
	//other difference is that n0[] is destroyed in this routine (speed hack).
	// Inputs: uind[MAXZDIM]: uncompressed bitbuf (0=air,1=solid)
	//         n0[MAXZDIM]: neighbor bitbuf (n0&n1&n2&n3): (0=air,1=solid)
	//         px,py: parameters for setting unexposed voxel colors
	//Outputs: cbuf[MAXCSIZ]: compressed output buffer
	//Returns: n: length of compressed buffer (in bytes)
long compilebitstack (long *uind, long *n0, char *cbuf, long px, long py)
{
	lpoint3d p;
	long cz0, cz1, uz0, uz1, cz, n, onext, stat;

	p.x = px; p.y = py;
	uz1 = -1; cz = 0; n = 0; onext = 0;
	while (1)
	{
		uz0 = dntil1(uind,uz1+1); if (uz0 >= zsiz) break;
		uz1 = dntil0(uind,uz0+1);
		if (uz1 < zsiz) n0[(uz1-1)>>5] &= ~(1<<(uz1-1)); else uz1 = zsiz;
		cz0 = uz0; stat = 0;
		while (1)
		{
			cz1 = dntil1(n0,cz0+1); if (cz1 > uz1) cz1 = uz1;
			if (!stat)
			{
				cbuf[onext] = ((n-onext)>>2); onext = n;
				cbuf[n+1] = cz0;
				cbuf[n+2] = cz1-1;
				cbuf[n+3] = cz;
				n += 4;
			}
			for(p.z=cz0;p.z<cz1;p.z++)
			{
				*(long *)&cbuf[n] = ((getpix32(p.x,p.y,p.z)&0xffffff)|0x80000000);
				n += 4;
			}
			if ((cz1 < uz1) && (stat))
			{
				cbuf[onext] = ((n-onext)>>2); onext = n;
				cbuf[n+1] = cz1;
				cbuf[n+2] = cz1-1;
				cbuf[n+3] = cz1;
				n += 4;
			}
			cz0 = dntil0(n0,cz1+1); if (cz0 >= uz1) break;
			stat = 1;
		}
		cz = uz1;
	}
	cbuf[onext] = 0;
	return(n);
}

#endif

lpoint2d vxledge[2];
void savevxl (char *fnam)
{
	FILE *fil;
	long i, x, y, z, *v, *v0, *v1;
	double d[3];
	long vbitemp[MAXZDIM>>5];

	if (!rendermode)
	{
			//Make sure bottom is solid
		for(x=0;x<VSID;x++)
			for(y=0;y<VSID;y++)
			{
#if 0          //Fill bottom up to lowest voxel
				i = (x*MAYDIM+y)*MAZDIM;
				z = uptil1(&vbit[i>>5],MAXZDIM-1);
				if (z) { setzrange1(&vbit[i>>5],z,256); }
				  else { i += MAXZDIM-1; vbit[i>>5] |= (1<<i); }
#else          //Write bottom voxel only
				i = (x*MAYDIM+y)*MAZDIM+MAXZDIM-1; vbit[i>>5] |= (1<<i);
#endif
			}

			//Make edges solid
		for(i=0;i<2;i++)
		{
			for(x=vxledge[0].x;x<=vxledge[1].x;x++)
				memset((void *)&vbit[((x*MAYDIM+vxledge[i].y)*MAZDIM)>>5],-1,MAXZDIM>>3);
			for(y=vxledge[0].y;y<=vxledge[1].y;y++)
				memset((void *)&vbit[((vxledge[i].x*MAYDIM+y)*MAZDIM)>>5],-1,MAXZDIM>>3);
		}
		hollowfix(xsiz,ysiz,zsiz);

		//NOTE: should clean up edges here, but how?
	}
	printf("Writing %s (%dx%dx%d)\n",fnam,xsiz,ysiz,zsiz);

	if (!(fil = fopen(fnam,"wb"))) quitout("Could not write file");

	i = 0x09072000; fwrite(&i,4,1,fil); //Version
	i = MAXDIM; fwrite(&i,4,1,fil);
	i = MAYDIM; fwrite(&i,4,1,fil);

		//Make sure starting camera is close to ground
	for(i=-5;issphereempty(512,512,i,5);i++);

	d[0] = 512.0; d[1] = 512.0; d[2] = (double)i; fwrite(d,24,1,fil);
	d[0] = 1.0; d[1] = 0.0; d[2] = 0.0; fwrite(d,24,1,fil);
	d[0] = 0.0; d[1] = 0.0; d[2] = 1.0; fwrite(d,24,1,fil);
	d[0] = 0.0; d[1] =-1.0; d[2] = 0.0; fwrite(d,24,1,fil);

		//Make sure bottom is solid
	for(x=0;x<VSID;x++)
		for(y=0;y<VSID;y++)
		{
#if 0       //Fill bottom up to lowest voxel
			i = (x*MAYDIM+y)*MAZDIM;
			z = uptil1(&vbit[i>>5],MAXZDIM-1);
			if (z) { setzrange1(&vbit[i>>5],z,256); }
			  else { i += MAXZDIM-1; vbit[i>>5] |= (1<<i); }
#else       //Write bottom voxel only
			i = (x*MAYDIM+y)*MAZDIM+MAXZDIM-1; vbit[i>>5] |= (1<<i);
#endif

		}

	for(x=0;x<VSID;x++)
		for(y=0;y<VSID;y++)
		{
			v = &vbit[((x*MAYDIM+y)*MAZDIM)>>5];

				//Collect the 4 neighbor columns (x-1,x+1,y-1,y+1) into vbitemp:
			v0 = &v[-((MAYDIM*MAZDIM)>>5)]; v1 = &v[(MAYDIM*MAZDIM)>>5];
			if (x == VSID-1) { memcpy(vbitemp,v0,sizeof(vbitemp)); }
			else if (!x)     { memcpy(vbitemp,v1,sizeof(vbitemp)); }
			else { for(z=(MAXZDIM>>5)-1;z>=0;z--) vbitemp[z] = (v0[z]&v1[z]); }

			v0 = &v[-(MAZDIM>>5)]; v1 = &v[MAZDIM>>5]; z = (MAXZDIM>>5)-1;
			if (y == VSID-1) { for(;z>=0;z--) vbitemp[z] &= v0[z]; }
			else if (!y)     { for(;z>=0;z--) vbitemp[z] &= v1[z]; }
			else             { for(;z>=0;z--) vbitemp[z] &= (v0[z]&v1[z]); }

			i = compilebitstack(v,vbitemp,tempbuf,x,y);
			fwrite(tempbuf,i,1,fil);
		}
	fclose(fil);
}

#if (PRINTTIME == 1)
static __forceinline __int64 rdtsc64 () { _asm rdtsc }
#endif

long loadpoly (char *ofnam)
{
	char fnam[MAX_PATH+4], *cptr;
	long i, fnlng, slash;

	strcpy(fnam,ofnam); fnlng = strlen(fnam);
	for(i=fnlng-1,slash=0;i>=0;i--) if ((fnam[i] == '\\') || (fnam[i] == '/')) { slash = i+1; break; }

	if ((fnlng >= 4) && (!strcmp(&fnam[fnlng-4],".zip"))) kzaddstack(fnam);
	else
	{
		if (fnlng >= 4)
		{
			cptr = &fnam[fnlng-4];
			if ((!stricmp(cptr,".md3")) && (loadmd3(fnam))) return(1);
			if ((!stricmp(cptr,".md2")) && (loadmd2(fnam))) return(1);
			if ((!stricmp(cptr,".3ds")) && (load3ds(fnam))) return(1);
			if ((!stricmp(cptr,".asc")) && (loadasc(fnam))) return(1);
		}
		strcpy(fnam,ofnam);
		strcpy(&fnam[fnlng],".md3"); if (loadmd3(fnam)) return(1);
		strcpy(&fnam[fnlng],".md2"); if (loadmd2(fnam)) return(1);
		strcpy(&fnam[fnlng],".3ds"); if (load3ds(fnam)) return(1);
		strcpy(&fnam[fnlng],".asc"); if (loadasc(fnam)) return(1);

		strcpy(&fnam[fnlng],".zip"); kzaddstack(fnam);
	}

	if (fnlng >= 4)
	{
		cptr = &fnam[fnlng-4];
		strcpy(cptr,".md3"); if ((loadmd3(fnam)) || (loadmd3(&fnam[slash]))) return(1);
		strcpy(cptr,".md2"); if ((loadmd2(fnam)) || (loadmd2(&fnam[slash]))) return(1);
		strcpy(cptr,".3ds"); if ((load3ds(fnam)) || (load3ds(&fnam[slash]))) return(1);
		strcpy(cptr,".asc"); if ((loadasc(fnam)) || (loadasc(&fnam[slash]))) return(1);
	}
	kzfindfilestart("*.*");
	while (kzfindfile(fnam))
	{
		fnlng = strlen(fnam); if ((fnlng < 4) || (fnam[0] != '|')) continue;
		cptr = &fnam[fnlng-4];
		if ((!stricmp(cptr,".md3")) && (loadmd3(fnam))) return(1);
		if ((!stricmp(cptr,".md2")) && (loadmd2(fnam))) return(1);
		if ((!stricmp(cptr,".3ds")) && (load3ds(fnam))) return(1);
		if ((!stricmp(cptr,".asc")) && (loadasc(fnam))) return(1);
	}
	return(0);
}

void showhelp ()
{
	printf("POLY2VOX [input] [output] [/v#] [/s#] [/f#] [/n#] [/r#] [/m#] [/x#] [/p(file)]\n");
	printf("by Ken Silverman (http://advsys.net/ken)  Compiled: %s\n\n",__DATE__);
	printf("Converts models from polygon to voxel format.\n");
	printf("Supported polygon formats: ASC,3DS,MD2,MD3\n");
	printf("Supported   voxel formats: VOX,KVX,KV6,VXL (default:KV6)\n");
	printf("Supported texture formats: PNG,JPG,TGA,GIF,CEL,PCX,BMP\n");
	printf("POLY2VOX can load files out of a ZIP file. (Model name must match ZIP name)\n\n");
	printf(" /v#  Specify voxel size of longest dimension. 1-1024, <=256 for KVX\n");
	printf(" /s#  Specify explicit scale factor. Use this to ensure the size of all frames\n");
	printf("      is consistent. This factor depends on the coordinate system used by the\n");
	printf("      polygon model, so it can be anything. Run without the scale factor first\n");
	printf("      to find a reasonble starting value to try.\n");
	printf(" /f#  Specify frame number (MD2/MD3 only)\n");
	printf(" /n#  Specify next frame number for interpolation (MD2/MD3 only)\n");
	printf(" /r#  Specify interpolation ratio. Range: 0.0-1.0, default:0.0 (MD2/MD3 only)\n");
	printf(" /m#  Specify number of mips to save: 1,5, default:5. (KVX only)\n");
	printf(" /x   Enable an experimental xor-style converter. It's useful for\n");
	printf("      gap-less models but has buggy color conversion\n");
	printf(" /p(file)  Specify a pre-defined palette (Build PALETTE.DAT format: first 768\n");
	printf("      bytes, range:0-63)\n\n");
	printf("Examples:\n");
	printf(" poly2vox bike                        (finds bike.*, writes bike.kvx, size=128)\n");
	printf(" poly2vox bike.3ds bike.kv6 /v250   (reads bike.3ds, writes bike.kv6, size=250)\n");
	printf(" poly2vox land land.vxl        (finds land.*, writes land.vxl, size=1024^2*256)\n");
	printf(" poly2vox pig.md2 pig.kvx /v128 /f2 /n3 /r.5       (convert interpolated frame)\n");
	printf(" poly2vox trooper.md2 trooper1680.kvx /s.115 /f0    (use explicit scale factor)\n");
	printf(" poly2vox trooper.md2 trooper1680.kvx /s0.115 /f0 /ppalette.dat\n\n");
}

void main (int argc, char **argv)
{
#if (PRINTTIME == 1)
	__int64 t0, t1;
#endif
	point3d oldp;
	float f, g, h, fx, fy, fz, fx0, fy0, fz0, fx1, fy1, fz1, forcescale = 0;
	long i, j, x, y, z, infilarg, outfilarg, voxdim = 128, numips = 5;
	char fnam[MAX_PATH], infilnam[MAX_PATH], outfilnam[MAX_PATH], *cptr;

#if (PRINTTIME == 1)
	t0 = rdtsc64();
#endif

	if (argc < 2) { showhelp(); exit(0); }
	infilarg = -1; outfilarg = -1;
	for(i=1;i<argc;i++)
	{
		if ((argv[i][0] != '-') && (argv[i][0] != '/'))
		{
				  if (infilarg < 0) infilarg = i;
			else if (outfilarg < 0) outfilarg = i;
			else { printf("ERROR: Too many files specified\n"); showhelp(); exit(0); }
			continue;
		}
		if ((argv[i][1] == 'V') || (argv[i][1] == 'v')) { voxdim = min(max(atol(&argv[i][2]),1),max(max(MAXDIM,MAYDIM),MAZDIM)); continue; }
		if ((argv[i][1] == 'S') || (argv[i][1] == 's')) { forcescale = atof(&argv[i][2]); continue; }
		if ((argv[i][1] == 'F') || (argv[i][1] == 'f')) { gcframe = atol(&argv[i][2]); continue; }
		if ((argv[i][1] == 'N') || (argv[i][1] == 'n')) { gnframe = atol(&argv[i][2]); continue; }
		if ((argv[i][1] == 'R') || (argv[i][1] == 'r')) { ginterp = atof(&argv[i][2]); continue; }
		if ((argv[i][1] == 'M') || (argv[i][1] == 'm')) { numips = atol(&argv[i][2]); if (numips <= 1) numips = 1; else numips = 5; continue; }
		if ((argv[i][1] == 'P') || (argv[i][1] == 'p')) { strcpy(palfilnam,&argv[i][2]); continue; }
		if ((argv[i][1] == 'X') || (argv[i][1] == 'x')) { rendermode = 1; continue; }
	}
	if (infilarg < 0) { printf("ERROR: Must specify input file\n"); showhelp(); exit(0); }

	if (argv[infilarg][0] == '\"') //Strip quotes
		  { strcpy(infilnam,&argv[infilarg][1]); infilnam[max(strlen(infilnam)-1,0)] = 0; }
	else { strcpy(infilnam,argv[infilarg]); }
	if (outfilarg >= 0)
	{
		if (argv[outfilarg][0] == '\"') //Strip quotes
			  { strcpy(outfilnam,&argv[outfilarg][1]); infilnam[max(strlen(infilnam)-1,0)] = 0; }
		else { strcpy(outfilnam,argv[outfilarg]); }
	}

	if (!( trioff = (    long *)malloc(sizeof(long)*maxobjects ))) quitout("malloc failed: trioff");
	if (!(vertoff = (    long *)malloc(sizeof(long)*maxobjects ))) quitout("malloc failed: vertoff");
	if (!( matstr = (    long *)malloc(sizeof(long)*maxmat     ))) quitout("malloc failed: matstr");
	if (!( colrgb = (    long *)malloc(sizeof(long)*maxmat     ))) quitout("malloc failed: colrgb");
	if (!( nambuf = (    char *)malloc(sizeof(char)*maxnamsiz  ))) quitout("malloc failed: nambuf");
	if (!(    pic = (  pictyp *)malloc(sizeof(pictyp)*maxpics  ))) quitout("malloc failed: pic");
	if (!(    tri = (  trityp *)malloc(sizeof(trityp)*maxtris  ))) quitout("malloc failed: tri");
	if (!(   vert = (  vertyp *)malloc(sizeof(vertyp)*maxverts ))) quitout("malloc failed: vert");
	if (!(hashdat = (lpoint3d *)malloc(sizeof(lpoint3d)*maxhash))) quitout("malloc failed: hashdat");

	if (!loadpoly(infilnam)) quitout("Can't find input file");

	savetype = 2;
	if (outfilarg >= 0)
	{
		strcpy(fnam,outfilnam); _strlwr(fnam);
			  if (strstr(fnam,".vox")) savetype = 0;
		else if (strstr(fnam,".kvx")) savetype = 1;
		else if (strstr(fnam,".kv6")) savetype = 2;
		else if (strstr(fnam,".vxl")) savetype = 3;
	}

	for(i=0;i<numpics;i++)
	{
		strcpy(fnam,&nambuf[pic[i].namoff]); _strlwr(fnam);
		kpzload(&nambuf[pic[i].namoff],&pic[i].f,&pic[i].p,&pic[i].x,&pic[i].y);
		if (!pic[i].f)
		{
			for(j=strlen(&nambuf[pic[i].namoff])-1;j>=0;j--)
				if ((nambuf[pic[i].namoff+j] == '\\') || (nambuf[pic[i].namoff+j] == '/'))
				{
					kpzload(&nambuf[pic[i].namoff+j+1],&pic[i].f,&pic[i].p,&pic[i].x,&pic[i].y);
					break;
				}
		}
		if (pic[i].f) { printf(" %s loaded successfully\n",&nambuf[pic[i].namoff]); picsloaded++; }
		else printf(" %s NOT FOUND or not supported format\n",&nambuf[pic[i].namoff]);
	}


		//TEMP HACK to swap x&y axes!!!
	//for(z=0;z<numverts;z++)
	//{
	//   oldp = *(point3d *)&vert[z];
	//   vert[z].x = oldp.y;
	//   vert[z].y =-oldp.x;
	//   vert[z].z = oldp.z;
	//}


// Scale data to specified resolution ---------------------------------------

	fx0 = 1e32f; fx1 = -1e32f;
	fy0 = 1e32f; fy1 = -1e32f;
	fz0 = 1e32f; fz1 = -1e32f;
	for(z=0;z<numverts;z++)
	{
		if (vert[z].x < fx0) fx0 = vert[z].x;
		if (vert[z].y < fy0) fy0 = vert[z].y;
		if (vert[z].z < fz0) fz0 = vert[z].z;
		if (vert[z].x > fx1) fx1 = vert[z].x;
		if (vert[z].y > fy1) fy1 = vert[z].y;
		if (vert[z].z > fz1) fz1 = vert[z].z;
	}

	if (savetype != 3)
	{
		if (forcescale > 0)
			f = forcescale;
		else
		{
			h = (MAYDIM-.1f)/(fx1-fx0); //Find maximum scale factor (g)
			f = (MAXDIM-.1f)/(fy1-fy0); if (f < h) h = f;
			f = (MAZDIM-.1f)/(fz1-fz0); if (f < h) h = f;
			g = fx1-fx0; i = 'x';
			if (fy1-fy0 > g) { g = fy1-fy0; i = 'y'; }
			if (fz1-fz0 > g) { g = fz1-fz0; i = 'z'; }
			f = ((float)voxdim-.1f) / g;
			if (f > h)
			{
				f = h; ftol(h*g,&voxdim);
				printf(" NOTE: %cdim limited to %d\n",i,voxdim);
			}
			if (voxdim <= 0) quitout("error: voxdim must be > 0");
		}
		ftol((fy1-fy0)*f+.5,&xsiz); if ((savetype < 2) && (xsiz > 256)) printf(" WARNING: xsiz > 256!");
		ftol((fx1-fx0)*f+.5,&ysiz); if ((savetype < 2) && (ysiz > 256)) printf(" WARNING: ysiz > 256!");
		ftol((fz1-fz0)*f+.5,&zsiz); if ((savetype < 2) && (zsiz > 256-savetype)) printf(" WARNING: zsiz > %d!",256-savetype);

		fy = fy0; fx = fx0; fz = fz1;
	}
	else
	{
		xsiz = VSID; ysiz = VSID; zsiz = MAXZDIM;
		if ((MAXDIM < xsiz) || (MAYDIM < ysiz) || (MAZDIM < zsiz))
			quitout("ERROR: compiled with maximum dimensions too low");

		f = (ysiz-.1f)/(fx1-fx0);
		h = (xsiz-.1f)/(fy1-fy0); if (h < f) f = h;
		h = (zsiz-.1f)/(fz1-fz0); if (h < f) f = h;

			//Hack to center object in converted space:
			//   (fy0-?)*f+.05 = ysiz - ((fy1-?)*f+.05)
			//   (fx0-?)*f+.05 = xsiz - ((fx1-?)*f+.05)
			//   (?-fz0)*f+.05 = zsiz - ((?-fz1)*f+.05)
		g = 1.0 / f;
		fy = ((fy0+fy1) - (ysiz-.05*2)*g)*.5;
		fx = ((fx0+fx1) - (xsiz-.05*2)*g)*.5;
		fz = ((zsiz-.05*2)*g + (fz0+fz1))*.5;

		vxledge[0].x = (long)((fy0-fy)*f+.05f); if (vxledge[0].x < 0) vxledge[0].x = 0;
		vxledge[0].y = (long)((fx0-fx)*f+.05f); if (vxledge[0].y < 0) vxledge[0].y = 0;
		vxledge[1].x = (long)((fy1-fy)*f+.05f); if (vxledge[1].x >= VSID) vxledge[1].x = VSID-1;
		vxledge[1].y = (long)((fx1-fx)*f+.05f); if (vxledge[1].y >= VSID) vxledge[1].y = VSID-1;
	}
	printf("Scale factor used (voxel/polygon units):%9.6f\n",f);

	for(z=0;z<numverts;z++)
	{
		oldp = *(point3d *)&vert[z];
		vert[z].x = (oldp.y-fy)*f+.05f;
		vert[z].y = (oldp.x-fx)*f+.05f;
		vert[z].z = (fz-oldp.z)*f+.05f;
	}
	oldp.x = oldp.y = oldp.z = 0.f;
	xpiv = (oldp.y-fy)*f+.05f;
	ypiv = (oldp.x-fx)*f+.05f;
	zpiv = (fz-oldp.z)*f+.05f;
	if (savetype == 1) zpiv = ((float)zsiz)*.5; //Ugly hack

// Render geometry to vbit[] and colors to sparse hash ----------------------

	memset(vbit,0,sizeof(vbit));
	inithash();

	if (!rendermode)
	{
		//for(i=numtris-1;i>=0;i--) renderline0tri(i); //Minimal wireframe
		//for(i=numtris-1;i>=0;i--) renderline1tri(i); //Supercover wireframe
		for(i=numtris-1;i>=0;i--) renderface0tri(i); //Minimal face fill (best!)
		//for(i=numtris-1;i>=0;i--) renderface1tri(i); //Supercover face fill
	}
	else
	{
		for(i=numtris-1;i>=0;i--) renderfilltri(i);
		renderfillfinish();
	}

	if (outfilarg < 0)
	{
		strcpy(fnam,infilnam);
		for(i=0,cptr=0;fnam[i];i++) if (fnam[i] == '.') cptr = &fnam[i]; //replace extension with kv6
		if (!cptr) cptr = &fnam[strlen(fnam)];
		strcpy(cptr,".kv6");

			//If EXE has full path, it probably means this was run from explorer (filename dragged onto exe).
			//In this case, it is better to save in the file's directory, and not the default user dir.
		for(i=0;1;i++)
		{
			if (argv[0][i] == 0) //a ':' indicates exe has full path
			{
				for(i=strlen(fnam)-1;i>=0;i--) //Strip off directory
					if ((fnam[i] == '\\') || (fnam[i] == '/')) break;
				savekv6(&fnam[i+1]); break;
			}
			if (argv[0][i] == ':') { savekv6(fnam); break; } //full path
		}
	}
	else
	{
		switch(savetype)
		{
			case 0: savevox(outfilnam); break;
			case 1: savekvx(outfilnam,numips); break;
			case 2: savekv6(outfilnam); break;
			case 3: savevxl(outfilnam); break;
		}
	}

	while (numpics > 0)
	{
		numpics--;
		if (pic[numpics].f) free((void *)pic[numpics].f);
	}

#if (PRINTTIME == 1)
	t1 = rdtsc64();
	printf("Elapsed clock cycles: %I64d\n",t1-t0);
#endif

	quitout("Done.");
}

#if 0
	long xsiz, ysiz, zsiz, xpivot, ypivot, zpivot;
	long xstart[xsiz+1];
	short xyoffs[xsiz][ysiz+1];
	char rawslabdata[?];

	//sptr = xstart[x] + xyoffs[x][y];   //start of slabs (x,y)
	//eptr = xstart[x] + xyoffs[x][y+1]; //end of slabs (x,y)
	char slabztop;
	char slabzleng;
	char slabbackfacecullinfo; //Low 6 bits used to tell which of the 6 faces are exposed.
	char col[slabzleng];       //The array of colors from top to bottom
#endif

#if 0
!endif
#endif
